from sgvb_psd.utils.sim_varma import SimVARMA
import numpy as np
from sgvb_psd.utils.tf_utils import set_seed
from sgvb_psd.psd_estimator import PSDEstimator
import matplotlib.pyplot as plt

plt.rcParams["font.family"] = "Arial"

sigma = np.array([[1.0, 0.9], [0.9, 1.0]])
varCoef = np.array([[[0.5, 0.0], [0.0, -0.3]], [[0.0, 0.0], [0.0, -0.5]]])
vmaCoef = np.array([[[1.0, 0.0], [0.0, 1.0]]])
n = 256
set_seed(0)
var = SimVARMA(
    n_samples=n, var_coeffs=varCoef, vma_coeffs=vmaCoef, sigma=sigma
)

estimator = PSDEstimator(
    x=var.data, fs=2 * np.pi, N_theta=35, N_samples=50, ntrain_map=50
)
estimator.run(lr=0.01)

TXT = "SGVB"

fig, axs = plt.subplots(2, 2, figsize=(1, 1), sharex=True)
axs = estimator.plot(
    plot_periodogram=False,
    tick_ln=0,
    xlims=[0, np.pi],
    off_symlog=False,
    add_axes_labels=False,
    add_channel_labels=False,
    alpha=0,
    ax=axs,
)
for i, ax in enumerate(axs.flatten()):
    ax.set_xticks([])
    ax.set_yticks([])

    # turn off all spines (turn off axes frame)
    for spine in ax.spines.values():
        spine.set_visible(False)

    ax.annotate(
        TXT[i],
        xy=(0.5, 0.5),
        xycoords="axes fraction",
        ha="center",
        va="center",
        fontsize=100,
        color="C0",
        fontweight="bold",
    )
plt.savefig("../_static/logo.png", bbox_inches="tight", dpi=300)
plt.savefig("../_static/logo_100.png", bbox_inches="tight", dpi=100)
plt.savefig("../_static/favicon.png", bbox_inches="tight", dpi=25)


from PIL import Image, ImageDraw


# Open the saved image
im = Image.open("../_static/favicon.png").convert("RGBA")

# Create a circular mask
size = im.size
mask = Image.new("L", size, 0)
draw = ImageDraw.Draw(mask)
draw.ellipse((0, 0) + size, fill=255)

# Apply the mask to the image
output = Image.new("RGBA", size)
output.paste(im, (0, 0), mask)

# Save the circular cropped image
output.save("../_static/favicon.png")


import h5py
import matplotlib.pyplot as plt
import numpy as np
import rpy2.robjects as ro
import rpy2.robjects.packages as rpackages
from rpy2.robjects import r

from sgvb_psd.postproc.plot_psd import (
    format_axes,
    plot_peridogram,
    plot_psd,
    plot_psdq,
)
from sgvb_psd.utils.periodogram import get_periodogram

RESULT = "et_data.h5"
DATA = dict(
    A="ET_caseA_noise.h5",
    B="ET_caseB_noise.h5",
    C="ET_caseB_noise.h5",  # Case C uses the same data as case B
)


class Case:
    def __init__(self, case):
        self.case = case
        self.result_fn = RESULT
        self.data_fn = DATA[case]
        self.psdq = self._load_estimated_psd_quantiles()
        self.true = self._load_true_psd()
        self.data = self._load_periodogram()
        print("Loaded case ", self)

    def _load_true_psd(self):
        with h5py.File(self.data_fn, "r") as f:
            freq = f["true_psd/freq"][:]
            idx = (freq > 5) & (freq < 128)
            psd = f["true_psd/psd"][:].T[idx]

            # PSD is currnelty just the XX, YY, ZZ components
            # make it a 3D array (nans for off-diagonal elements)
            psd = np.array([np.diag(psd[i]) for i in range(psd.shape[0])])
            return [psd, freq[idx]]

    def _load_estimated_psd_quantiles(self):
        with h5py.File(self.result_fn, "r") as f:
            return [f[f"case{self.case}_psd_quantiles"][:], f[f"freq"][:]]

    def _load_periodogram(self):
        with h5py.File(self.data_fn, "r") as f:
            freq = f["periodogram/freq"][:]
            idx = (freq > 5) & (freq < 128)
            return [f["periodogram/pdgrm"][:][idx], freq[idx]]

    def __repr__(self):
        return (
            f"Case {self.case} ["
            f"psdq: {self.psdq[0].shape}, "
            f"true: {self.true[0].shape}, "
            f"data: {self.data[0].shape}]"
        )

    def plot(self, axes=None, fname=None, **kwargs):
        axes = plot_psd(
            self.psdq,
            self.true,
            self.data,
            axes=axes,
            xlims=[5, 128],
            diag_ylims=[1e-52, 1e-46],
            off_ylims=[-3e-47, 3e-47],
            **kwargs,
        )
        if fname is not None:
            plt.savefig(fname)
            print(f"Saved {fname}")
        return axes


def load_vnpc_results(fpath="mcmc_results.RData"):
    # Import the necessary R packages
    vnpc_avg = rpackages.importr("vnpc.avg")
    beyond_whittle = rpackages.importr("beyondWhittle")

    r["load"](fpath)
    r_obj = r["mcmc_vnp_avg"]
    result_dict = {}

    # Iterate over the ListVector items (names and values)
    for name, value in zip(r_obj.names, r_obj):
        # Convert R objects to appropriate Python/Numpy structures
        if isinstance(value, ro.vectors.FloatMatrix):
            result_dict[name] = np.array(value)
        elif isinstance(value, ro.vectors.FloatArray):
            result_dict[name] = np.array(value)
        else:
            result_dict[name] = value  # Keep as it is for other data types

    coh = np.array(
        [
            result_dict["coherence.p05"],
            result_dict["coherence.median"],
            result_dict["coherence.p95"],
        ]
    )
    data = result_dict["data"]
    psd = np.array(
        [  # pointwise
            result_dict["psd.p05"],
            result_dict["psd.median"],
            result_dict["psd.p95"],
        ]
    )

    x, d, d, nfreq = psd.shape
    # reshape to x, nfreq. d, d
    psd = np.transpose(psd, (0, 3, 1, 2))
    # cut 0th and last nfreq
    psd = psd[:, 1:-1, :, :]

    psd = psd.astype(np.complex128)
    # make lower triangle elements imaginary by multiplying by 1j
    for i in range(d):
        for j in range(d):
            if i > j:
                psd[:, :, i, j] *= 1j

    # rescale PSDs (1e+22 is approx std for data)
    psd = psd / 1e22**2 / (1024.0 / np.pi)
    freq = np.linspace(5, 128, nfreq)

    return psd, freq


if __name__ == "__main__":
    mcmc_psd, mcmc_freq = load_vnpc_results("et_mcmc.RData")
    axes = Case("A").plot(color="C1")
    axes = plot_psdq(mcmc_psd, freqs=mcmc_freq, color="tab:red", axes=axes)
    plt.savefig("mcmc_et.png", dpi=500)



from click.testing import CliRunner
from sgvb_psd.utils.benchmark.cli import cli
import os

def test_benchmark(plot_dir):
    runner = CliRunner()
    result = runner.invoke(cli, ['-s', '4', '-e', '5', '-f', f'{plot_dir}/timings.txt'])
    assert result.exit_code == 0
    assert os.path.exists(f'{plot_dir}/timings.txt')



import os
from collections import namedtuple

import numpy as np
import pytest
import tensorflow as tf

HERE = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def output_dir():
    d = os.path.join(HERE, "out")
    os.makedirs(plot_dir, exist_ok=True)
    return d



@pytest.fixture
def plot_dir():
    plot_dir = os.path.join(HERE, "out_plots")
    os.makedirs(plot_dir, exist_ok=True)
    return plot_dir


@pytest.fixture
def et_data():
    # download and cache the dataset (small) if it doesn't exist
    pass



import logging
import os

import emcee
import matplotlib.pyplot as plt
import numpy as np
from tqdm.auto import trange

from sgvb_psd.postproc import plot_psdq
from sgvb_psd.psd_estimator import PSDEstimator
from sgvb_psd.utils.sim_varma import SimVARMA

# Set logging level
logging.getLogger("SGVB-PSD").setLevel("ERROR")

# Constants
N_EXPERIMENTS = 10
FS = 100
OUTDIR = "out_gibbs"
os.makedirs(OUTDIR, exist_ok=True)

# Simulation and VI configurations
SIM_KWGS = dict(
    sigma=np.array([[1.0, 0.9], [0.9, 1.0]]),
    var_coeffs=np.array(
        [[[0.5, 0.0], [0.0, -0.3]], [[0.0, 0.0], [0.0, -0.5]]]
    ),
    vma_coeffs=np.array([[[1.0, 0.0], [0.0, 1.0]]]),
    n_samples=1024,
)

VI_KWGS = dict(
    N_theta=30,
    nchunks=1,
    ntrain_map=1000,
    max_hyperparm_eval=1,
    fs=FS,
    seed=0,
    N_samples=1,
)


# Signal model function
def signal_model(t: np.ndarray, f0: float, a: float = 5.0):
    y = a * np.sin(2 * np.pi * f0 * t)
    return np.array([y, y]).T


# Plot simulation PSD
def plot_simulation_psd(optim: PSDEstimator, sim: SimVARMA):
    axs = optim.plot(
        true_psd=[sim.psd, sim.freq],
        off_symlog=False,
        xlims=[0, FS / 2.0],
        quantiles="pointwise",
    )
    plot_psdq(
        optim.uniform_ci,
        freqs=optim.freq,
        axs=axs,
        color="red",
        ls="--",
    )
    return axs


# Utility functions
def get_univar_psd(psd_matrix):
    return np.real(np.array([psd_matrix[:, 0, 0], psd_matrix[:, 1, 1]])).T


def log_prob(theta, xf, psd):
    f0 = theta[0]
    if not 0 < f0 < FS / 2:
        return -np.inf

    signal = signal_model(t, f0)
    signal_fft = np.fft.rfft(signal, axis=0)
    residual = xf - signal_fft
    residual = residual[1:-1]
    return -0.5 * np.sum(residual**2 / psd)


# Run MCMC
def run_mcmc(data, psd, x0=None):
    fft_data = np.fft.rfft(data, axis=0)
    univar_psd = get_univar_psd(psd)
    sampler = emcee.EnsembleSampler(
        nwalkers, ndim, log_prob, args=(fft_data, univar_psd)
    )
    if x0 is None:
        x0 = np.random.normal(true_f0, 1, size=(nwalkers, ndim))

    assert x0.shape == (nwalkers, ndim), x0.shape
    sampler.run_mcmc(x0, nsteps)
    samples = sampler.get_chain(discard=int(nsteps * 0.5), flat=True)
    x0 = sampler.get_chain()[-1, :, :]
    return x0, samples


# Run SGVB
def run_sgvb(data, x0, psd_x0=None):
    x0_mean = np.mean(x0)
    signal = signal_model(t, x0_mean)
    optim = PSDEstimator(**VI_KWGS, x=data - signal, init_params=psd_x0)
    optim.run(lr=0.003)
    psd_x0, psd = optim.sample_posterior(1)
    psd = psd[0][0]
    return psd_x0, psd


# Plot 1D PSD
def plot_1d(psd, x0, data, idx=None):
    plt.close("all")
    psd = get_univar_psd(psd).T[0]
    x0_mean = np.mean(x0)
    signal = signal_model(t, x0_mean)
    n = len(data)
    fft_signal = np.fft.rfft(signal.T[0], axis=0)[1:-1]
    fft_data = np.fft.rfft(data.T[0], axis=0)[1:-1]
    freq = np.fft.rfftfreq(n, d=1 / FS)[1:-1]

    plt.plot(freq, np.abs(fft_data), color="gray", label="data")
    plt.loglog(freq, psd, color="red", label="psd")
    plt.plot(freq, np.abs(fft_signal), color="blue", ls="--", label="signal")
    plt.legend()
    plt.xlabel("Frequency (Hz)")
    if idx is not None:
        plt.title(f"Iteration {idx}")
    plt.tight_layout()
    plt.savefig(f"{OUTDIR}/psd.png")


def plot_gibbs_results(fname):
    dataset = np.load(fname)

    x0s = dataset["x0s"]
    psds = dataset["psds"]
    data = dataset["data"]

    n_psds = psds.shape[0]

    # lets grab last 10% of samples
    n_psds = int(n_psds * 0.1)

    # just grab n_psd samples for x0s
    x0s = x0s[:n_psds]

    signals = np.array([signal_model(t, x0) for x0 in x0s])[..., 0]
    signal_fft = np.abs(np.fft.rfft(signals, axis=1))[..., 1:-1]
    data_fft = np.abs(np.fft.rfft(data, axis=0))[1:-1]
    freq = np.fft.rfftfreq(len(data), d=1 / FS)[1:-1]

    signal_fft_ci = np.percentile(signal_fft, [5.0, 95.0], axis=0)
    psd_ci = np.percentile(np.real(psds), [5.0, 95.0], axis=0)

    fig, axes = plt.subplots(1, 2, figsize=(8, 6))
    ax = axes[0]
    all_x0s = dataset["x0s"]
    # keep only last 10% of samples
    all_x0s = all_x0s[int(all_x0s.shape[0] * 0.9) :]
    ax.hist(all_x0s, color="gray", alpha=0.5)
    ax.axvline(true_f0, color="red", ls="--")
    ax.set_xlabel("f0 samples")

    ax = axes[1]
    ax.loglog(freq, data_fft, color="gray", label="data")
    ax.fill_between(
        freq, psd_ci[0], psd_ci[1], color="red", alpha=0.3, label="psd"
    )
    ax.fill_between(
        freq,
        signal_fft_ci[0],
        signal_fft_ci[1],
        color="blue",
        alpha=0.3,
        label="signal",
    )
    ax.legend()
    ax.set_xlabel("Frequency (Hz)")
    plt.tight_layout()
    plt.savefig(fname.replace(".npz", ".png"))


# Initialize simulation
sim = SimVARMA(**SIM_KWGS, seed=0)
dt = 1 / FS
t = np.linspace(0, (sim.n_samples - 1) * dt, sim.n_samples)
true_f0 = 5.0
true_signal = signal_model(t, true_f0)
sim.data += true_signal

# MCMC parameters
nwalkers = 16
ndim = 1
nsteps = 1000


def main():
    # Gibbs sampling
    gibbs = 500
    x0 = np.random.normal(true_f0, 1, size=(nwalkers, ndim))
    x0s = []
    psds = []
    psd_x0 = None

    for i in trange(gibbs, desc="Gibbs sampling"):
        x0, samples = run_mcmc(sim.data, sim.psd, x0)
        psd_x0, psd = run_sgvb(sim.data, x0, psd_x0)
        x0s.append(x0)
        psds.append(psd[:, 0, 0])
        if i % 50 == 0 and i > 0:
            plot_1d(psd, x0, sim.data, i)

    x0s = np.array(x0s).ravel()
    psds = np.array(psds)
    np.savez(f"{OUTDIR}/gibbs.npz", x0s=x0s, psds=psds, data=sim.data.T[0])

    plot_gibbs_results(f"{OUTDIR}/gibbs.npz")


if __name__ == "__main__":
    main()



import logging
import sys
import time

import tensorflow as tf
from colorama import Fore, Style, init


def log_if_gpu_or_cpu():
    gpus = tf.config.list_physical_devices("GPU")
    if gpus:
        logger.info(f"GPU found: {gpus}")
    else:
        logger.info("No GPU detected. Running on CPU.")


# Initialize colorama
init(autoreset=True)


class RelativeSecondsColoredFormatter(logging.Formatter):
    COLORS = {
        "DEBUG": Fore.BLUE,
        "INFO": Fore.GREEN,
        "WARNING": Fore.YELLOW,
        "ERROR": Fore.RED,
        "CRITICAL": Fore.RED + Style.BRIGHT,
    }

    def __init__(self, fmt=None, datefmt=None, style="%"):
        super().__init__(fmt, datefmt, style)
        self.start_time = time.time()

    def format(self, record):
        # Calculate relative seconds
        relative_seconds = int(record.created - self.start_time)
        record.relativeSeconds = relative_seconds

        # Apply color formatting
        log_color = self.COLORS.get(record.levelname, "")
        message = super().format(record)
        return f"{log_color}{message}{Style.RESET_ALL}"


def setup_logger(name, level=logging.INFO):
    # Create a custom logger
    logger = logging.getLogger(name)
    logger.setLevel(level)  # Set the default logging level

    # Create a console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.DEBUG)

    # Define the log message format

    # log_format = "%(asctime)s|%(name)s|%(levelname)s| %(relativeSeconds)ds |%(message)s"
    log_format = "%(asctime)s|%(name)s|%(levelname)s| %(message)s"

    formatter = RelativeSecondsColoredFormatter(log_format, datefmt="%H:%M:%S")
    console_handler.setFormatter(formatter)

    # Add the handler to the logger
    logger.addHandler(console_handler)

    return logger


logger = setup_logger("SGVB-PSD")
log_if_gpu_or_cpu()



import os.path
import time

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from sgvb_psd.backend import BayesianModel, AnalysisData
from sgvb_psd.logging import logger
from sgvb_psd.postproc.plot_psd import plot_psdq
from sgvb_psd.postproc.psd_analyzer import PSDAnalyzer
from sgvb_psd.psd_estimator import PSDEstimator
from sgvb_psd.utils.sim_varma import SimVARMA
from sgvb_psd.utils.tf_utils import set_seed


def get_var_data(npts):
    set_seed(0)
    tf.config.experimental.enable_op_determinism()

    sigma = np.array([[1.0, 0.9], [0.9, 1.0]])
    varCoef = np.array([[[0.5, 0.0], [0.0, -0.3]], [[0.0, 0.0], [0.0, -0.5]]])
    vmaCoef = np.array([[[1.0, 0.0], [0.0, 1.0]]])
    print(f"Generating VARMA data: {npts} samples")
    var2 = SimVARMA(
        n_samples=npts, var_coeffs=varCoef, vma_coeffs=vmaCoef, sigma=sigma
    )
    return var2


def test_var_psd_generation(plot_dir):
    var2_data = get_var_data(2 ** 9)
    print("Starting VAR PSD generation test")
    start_time = time.time()
    kwags = dict(
        N_theta=30,
        nchunks=1,
        ntrain_map=100,
        x=var2_data.data,
        max_hyperparm_eval=1,
        n_elbo_maximisation_steps=50,
        fs=2 * np.pi,
        seed=0,)
    optim = PSDEstimator(**kwags)

    optim.run()

    ## Run is done, lets make some plots
    ax = optim.plot(
        true_psd=[var2_data.psd, var2_data.freq],
        off_symlog=False,
        xlims=[0, np.pi],
        quantiles='pointwise',
    )
    plot_psdq(
        optim.uniform_ci,
        freqs=optim.freq,
        axs=ax,
        color='red',
        ls='--',
    )

    plt.savefig(f"{plot_dir}/var_psd250.png")
    optim.plot_coherence(
        true_psd=[var2_data.psd, var2_data.freq],
    )
    plt.savefig(f"{plot_dir}/var_coherence.png")
    optim.plot_vi_losses()
    plt.savefig(f"{plot_dir}/var_vi_losses.png")
    csv = f"{plot_dir}/var_psd.csv"
    psd_analyzer = PSDAnalyzer(
        spec_true=var2_data.psd,
        spectral_density_q=optim.pointwise_ci,
        idx=1,
        csv_file=csv,
    )
    assert isinstance(psd_analyzer.coverage_point_CI, float)
    assert isinstance(psd_analyzer.l2_error, float)
    assert os.path.exists(csv)

    end_time = time.time()
    estimation_time = end_time - start_time
    assert estimation_time < 50, f"Estimation time {estimation_time} is too long"
    logger.info(f"Test passed in {estimation_time} seconds")

    # see if we can load old params
    spline_params, psd = optim.sample_posterior(1)
    optim2 = PSDEstimator(**kwags, init_params=spline_params)



def test_one_train_step_with_chunks():
    t0 = time.time()
    var2_data = get_var_data(2 ** 7)
    analysis_data = AnalysisData(
        x=var2_data.data,
        nchunks=2,
        fmax_for_analysis=128,
        fs=2 * np.pi,
        N_theta=10,
        N_delta=10,
    )
    model = BayesianModel(analysis_data)
    opts = tf.keras.optimizers.Adam(0.01)
    model.map_train_step(opts)
    estimation_time = time.time() - t0
    assert estimation_time < 2, f"Estimation time {estimation_time} is too long"
    logger.info(f"Test passed in {estimation_time} seconds")

#
# def test_var_psd_save_and_resume(plot_dir, output_dir):
#     var2_data = get_var_data(2 ** 8)
#     optim = PSDEstimator(
#         N_theta=30,
#         nchunks=1,
#         ntrain_map=100,
#         max_hyperparm_eval=1,
#         n_elbo_maximisation_steps=50,
#         fs=2 * np.pi,
#         seed=0,
#     )
#     optim.run(x=var2_data.data)
#     optim.save(output_dir)
#
#     optim = PSDEstimator.load(output_dir)
#     optim.run(x=var2_data.data)



import time
import timeit
from typing import List, Tuple

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow.keras.optimizers import Adam

from ..logging import logger
from .analysis_data import AnalysisData
from .bayesian_model import BayesianModel

tfd = tfp.distributions
tfb = tfp.bijectors


class ViRunner:
    def __init__(
        self,
        x: np.ndarray,
        N_theta: int = 30,
        nchunks: int = 400,
        variation_factor: float = 0.0,
        fmax_for_analysis: float = None,
        fs: float = 2048,
        degree_fluctuate: float = None,
        init_params: List[tf.Tensor] = None,
        surrogate_posterior: tfd.JointDistributionSequential = None,
        fmin_for_analysis: float = None,
    ):
        self.data = AnalysisData(
            x=x,
            nchunks=nchunks,
            fmax_for_analysis=fmax_for_analysis,
            fs=fs,
            N_theta=N_theta,
            N_delta=N_theta,  # N_theta == N_delta in all cases
            fmin_for_analysis=fmin_for_analysis,
        )

        ## Define Model
        self.model = BayesianModel(
            self.data,
            degree_fluctuate=degree_fluctuate,
            init_params=init_params,
        )
        self.variation_factor = variation_factor
        if surrogate_posterior is not None:
            self.surrogate_posterior = surrogate_posterior
        else:
            self.surrogate_posterior: tfd.JointDistributionSequential = None

    def run(
        self,
        lr_map: float = 5e-4,
        ntrain_map: int = 5000,
        inference_size: int = 500,
        n_elbo_maximisation_steps: int = 1000,
    ) -> Tuple[np.ndarray, np.ndarray, BayesianModel, List[tf.Tensor]]:
        logger.debug("Starting Model Inference Training..")
        self.run_phase1(lr_map, ntrain_map)
        self.run_phase2(n_elbo_maximisation_steps)
        # now that the model is trained, we can sample from it
        samp = self.surrogate_posterior.sample(inference_size)

        return self.kdl_losses, self.lp, self.model, samp

    def run_phase1(
        self,
        lr_map: float = 5e-4,
        ntrain_map: int = 5000,
    ):
        """set model.trainable_vars to MAP values

        # Note: at this point, the model.trainable_vars are set to
        # param values that maximise the log posterior
        # THESE ARE NO LONGER CHANGED AFTER PHASE 1
        """
        optimizer_hs = Adam(lr_map)
        start_map = timeit.default_timer()
        logger.debug(f"Start Phase 1: MAP search ({ntrain_map} steps)... ")
        ntrain_map = tf.constant(ntrain_map, dtype=tf.int32)

        @tf.function(reduce_retracing=True)
        def tune_model_to_map(
            model: BayesianModel, optimizer: Adam, n_train: int
        ) -> Tuple[List[tf.Variable], tf.Tensor]:
            n_samp = model.trainable_vars[0].shape[0]
            lpost = tf.constant(0.0, tf.float32, [n_samp])
            lp = tf.TensorArray(tf.float32, size=0, dynamic_size=True)
            for i in tf.range(n_train):
                lpost = model.map_train_step(optimizer)

                if optimizer.iterations % 5000 == 0:
                    tf.print(
                        "Step",
                        optimizer.iterations,
                        "/",
                        n_train,
                        " : log posterior",
                        lpost,
                    )
                lp = lp.write(tf.cast(i, tf.int32), lpost)

            return model.trainable_vars, lp.stack()

        opt_vars_hs, self.lp = tune_model_to_map(
            self.model, optimizer_hs, ntrain_map
        )
        self.map_time = timeit.default_timer() - start_map
        logger.debug(f"MAP Training Time: {self.map_time:.2f}s")

    def run_phase2(
        self,
        n_elbo_maximisation_steps: int = 1000,
    ):
        """
        Phase 2 UQ
        """
        optimizer_vi = Adam(5e-2)

        self.init_surrogate_posterior(params=self.model.trainable_vars)

        def conditioned_log_prob(*z):
            return self.model.loglik(z) + self.model.logprior(z)

        logger.debug(
            f"Start Phase 2: ELBO maximisation ({n_elbo_maximisation_steps} steps)... "
        )
        start_vi = timeit.default_timer()
        # For more on TF's fit_surrogate_posterior, see
        # https://www.tensorflow.org/probability/api_docs/python/tfp/vi/fit_surrogate_posterior
        self.kdl_losses = tf.function(
            lambda l: tfp.vi.fit_surrogate_posterior(
                target_log_prob_fn=l,
                surrogate_posterior=self.surrogate_posterior,
                optimizer=optimizer_vi,
                num_steps=n_elbo_maximisation_steps,
                # jit_compile=True, ## this is running slower than without it
            )
        )(
            conditioned_log_prob
        )  #
        self.vi_time = timeit.default_timer() - start_vi
        logger.debug(f"VI Time: {self.vi_time:.2f}s")
        self.total_time = self.map_time + self.vi_time
        logger.debug(f"Total Inference Training Time: {self.total_time:.2f}s")

        self.posteriorPointEst = self.surrogate_posterior.mean()
        self.posteriorPointEstStd = self.surrogate_posterior.stddev()
        self.variationalDistribution = self.surrogate_posterior

    def init_surrogate_posterior(self, params: List[tf.Variable]) -> None:
        if self.variation_factor <= 0:
            self.surrogate_posterior = tfd.JointDistributionSequential(
                [
                    tfd.Independent(
                        tfd.MultivariateNormalDiag(
                            loc=params[i][0],
                            scale_diag=tfp.util.TransformedVariable(
                                tf.constant(
                                    1e-4, tf.float32, params[i][0].shape
                                ),
                                tfb.Softplus(),
                                name="q_z_scale",
                            ),
                        ),
                        reinterpreted_batch_ndims=1,
                    )
                    for i in tf.range(len(params))
                ]
            )
        else:  # variation_factor > 0
            self.surrogate_posterior = tfd.JointDistributionSequential(
                [
                    tfd.Independent(
                        tfd.MultivariateNormalDiagPlusLowRank(
                            loc=params[i][0],
                            scale_diag=tfp.util.TransformedVariable(
                                tf.constant(
                                    1e-4, tf.float32, params[i][0].shape
                                ),
                                tfb.Softplus(),
                            ),
                            scale_perturb_factor=tfp.util.TransformedVariable(
                                tf.random_uniform_initializer()(
                                    params[i][0].shape + self.variation_factor
                                ),
                                tfb.Identity(),
                            ),
                        ),
                        reinterpreted_batch_ndims=1,
                    )
                    for i in tf.range(len(params))
                ]
            )



from pathlib import Path

import matplotlib


def set_rcparms():
    # Absolute path to the top level of the repository
    rc_file = Path(__file__).resolve().absolute().parents[0] / "matplotlibrc"
    matplotlib.rc_file(rc_file)


set_rcparms()

from .plot_coherence import plot_coherence
from .plot_psd import format_axes, plot_peridogram, plot_psdq, plot_single_psd
from .psd_analyzer import PSDAnalyzer



from typing import Tuple

import numpy as np
from scipy import signal
from scipy.signal import csd


def get_periodogram(x, fs, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
    n, p = x.shape
    x = x
    periodogram = np.zeros((n // 2, p, p), dtype=complex)
    for row_i in range(p):
        for col_j in range(p):
            if row_i == col_j:
                f, Pxx_den0 = signal.periodogram(
                    x[:, row_i], fs=fs, scaling="density"
                )
                f = f[1:]
                Pxx_den0 = Pxx_den0[1:]
                periodogram[..., row_i, col_j] = Pxx_den0 / 2
            else:

                y = np.apply_along_axis(np.fft.fft, 0, x)
                if np.mod(n, 2) == 0:
                    # n is even
                    y = y[0 : int(n / 2)]
                else:
                    # n is odd
                    y = y[0 : int((n - 1) / 2)]
                y = y / np.sqrt(n)
                cross_spectrum_fij = y[:, row_i] * np.conj(y[:, col_j])
                cross_spectrum_fij = cross_spectrum_fij / fs
                periodogram[..., row_i, col_j] = cross_spectrum_fij

    return periodogram, f


def get_welch_periodogram(x, fs, n_chunks=1) -> Tuple[np.ndarray, np.ndarray]:
    """Given a multivariate time series, return the periodogram."""
    n, p = x.shape
    nperseg = n // n_chunks
    n, p = x.shape
    n_freq = nperseg // 2
    periodogram = np.zeros((n_freq, p, p), dtype=complex)
    for row_i in range(p):
        for col_j in range(p):
            xi = x[:, row_i]
            xj = x[:, col_j]
            f, Pxx_den0 = signal.csd(
                xi,
                xj,
                fs=fs,
                nperseg=nperseg,
                return_onesided=True,
                scaling="density",
                average="median",
                detrend=False,
            )
            periodogram[..., row_i, col_j] = Pxx_den0[1:]
    return periodogram, f[1:]



import argparse

import h5py
import matplotlib.pyplot as plt
import numpy as np

from sgvb_psd.psd_estimator import PSDEstimator

DATA_FILES = dict(
    caseA="ET-CaseA-noise.h5",
    caseB="ET-CaseB-noise.h5",
)
CHANNELS = "XYZ"


class Data:
    def __init__(self, case):
        self.case = case

        fname = DATA_FILES[case]
        with h5py.File(fname, "r") as f:
            self.freq = f["freq"][:]
            self.channels = np.array([f[c][:] for c in CHANNELS])
            self.true_psd = f["true_psd"][:]
            self.true_freq = None  # load this jianan -- idk how 'true_freq' is stored in the file


def run_analysis(case, label=""):
    data = Data(case)
    psd_estimator = PSDEstimator(
        x=data.channels,
        N_theta=400,
        nchunks=125,
        ntrain_map=1000,
        N_samples=500,
        fs=2048,
        max_hyperparm_eval=1,
        fmax_for_analysis=128,
        degree_fluctuate=400,
    )
    # we previously optimized the learning rate, now hardcoding it
    psd_estimator.run(lr=0.0189)
    # save the PSD quantiles
    with h5py.File(f"ET-{case}{label}-SGVB-PSD.h5", "w") as f:
        f["freq"] = psd_estimator.freq
        f["psd_pointwise_ci"] = psd_estimator.pointwise_ci
        f["psd_uniform_ci"] = psd_estimator.uniform_ci
    psd_estimator.plot(
        true_psd=[data.true_psd, data.true_freq],
        xlims=[5, 128],
    )
    plt.savefig(f"ET-{case}{label}-SGVB-PSD.png")


def combine_results():
    # load each PSD-quantile file
    psd_pointwise = {}
    psd_uniform = {}
    for case in DATA_FILES:
        with h5py.File(f"ET-{case}-SGVB-PSD.h5", "r") as f:
            psd_pointwise[case] = f["psd_pointwise_ci"][:]
            psd_uniform[case] = f["psd_uniform_ci"][:]
            freq = f["freq"][:]

    fpath = "ET-Case-ABC-SGVB-PSD.h5"
    print(f"Creating file: {fpath}")
    with h5py.File(fpath, "w") as f:
        for case in DATA_FILES:
            f.create_dataset(
                f"{case}/psd_pointwise_ci", data=psd_pointwise[case]
            )
            f.create_dataset(f"{case}/psd_uniform_ci", data=psd_uniform[case])
        f["freq"] = freq
    print(f"Data saved to: {fpath}")


def main():
    parser = argparse.ArgumentParser(
        description="Run PSD analysis for a specified case."
    )
    parser.add_argument(
        "--case",
        choices=DATA_FILES.keys(),
        required=True,
        help="Specify the case to analyze (caseA or caseB).",
    )
    parser.add_argument(
        "--label", type=str, required=True, help="Extra label for result."
    )

    args = parser.parse_args()

    run_analysis(args.case, args.output)
    # combine_results()  # Uncomment this if you want to combine results after analysis


if __name__ == "__main__":
    main()



from sgvb_psd.utils.sim_varma import SimVARMA
from sgvb_psd.utils.tf_utils import set_seed
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from sgvb_psd.utils.periodogram import get_welch_periodogram


def test_data_generation(plot_dir):
    sigma = np.array([[1.0, 0.9], [0.9, 1.0]])
    varCoef = np.array([[[0.5, 0.0], [0.0, -0.3]], [[0.0, 0.0], [0.0, -0.5]]])
    vmaCoef = np.array([[[1.0, 0.0], [0.0, 1.0]]])
    n = 4096
    set_seed(0)
    var = SimVARMA(
        n_samples=n, var_coeffs=varCoef, vma_coeffs=vmaCoef, sigma=sigma
    )

    set_seed(0)
    var2 = SimVARMA(
        n_samples=n, var_coeffs=varCoef, vma_coeffs=vmaCoef, sigma=sigma
    )

    # Gridspec
    # 4 colums and 2 rows
    # 2nd column is blank
    # arrange the plots in the following order
    # ax0     ax[0,0], ax[0,1],
    # ax1     ax[1,0], ax[1,1]

    fig = plt.figure(figsize=(8, 4))
    gs = gridspec.GridSpec(2, 4)
    ax0 = fig.add_subplot(gs[0, 0])
    ax1 = fig.add_subplot(gs[1, 0])

    ax02 = fig.add_subplot(gs[0, 2])
    ax03 = fig.add_subplot(gs[0, 3])
    ax12 = fig.add_subplot(gs[1, 2])
    ax13 = fig.add_subplot(gs[1, 3])

    axes = np.array([[ax02, ax03], [ax12, ax13]])

    ax0.plot(var.data[:, 0])
    ax1.plot(var.data[:, 1])

    assert np.all(var.data == var2.data)
    var2.plot(
        axs=axes, add_axes_labels=False, xlims=[0, np.pi], welch_nchunks=8
    )

    for ax in [ax12, ax13]:
        ax.set_xlabel("Frequency [Hz]", fontsize=12)
    for ax in [ax02, ax12]:
        ax.set_ylabel("PSD [1/Hz]", fontsize=12)

    ax1.set_xlabel("Time [s]", fontsize=12)
    for ax in [ax0, ax1]:
        ax.set_ylabel("Amplitude", fontsize=12)
        ax.set_xlim([0, n])

    # draw an arrow from [ax0, ax1] to [ax02, ax03]

    # add annotationn to top right of ax0, ax1
    ax0.annotate("Channel 1", xy=(0.6, 0.9), xycoords="axes fraction")
    ax1.annotate("Channel 2", xy=(0.6, 0.9), xycoords="axes fraction")
    ax0.set_xticklabels([])

    # Draw an arrow from ax0 (bottom right corner) to ax02 (bottom left corner)
    ax0.annotate(
        "",
        xy=(0.4, 0.5),  # end point
        xytext=(0.3, 0.5),  # start point
        xycoords="figure fraction",
        arrowprops=dict(arrowstyle="->", color="black", lw=1.5),
    )

    plt.savefig(f"{plot_dir}/pdgrm.png")

    assert var._repr_html_() is not None



import numpy as np


def get_freq(
    fs: float, n_time_samples: int, fmax=None, fmin=None
) -> np.ndarray:
    n = n_time_samples
    dt = 1 / fs
    freq = np.fft.fftfreq(n, d=dt)
    if np.mod(n, 2) == 0:  # the length per chunk is even
        freq = freq[1 : int(n / 2)]
    else:  # the length per chunk is odd
        freq = freq[1 : int((n - 1) / 2)]

    if fmax is not None:
        fmax_idx = np.searchsorted(freq, fmax)
        freq = freq[0:fmax_idx]

    if fmin is not None:
        fmin_idx = np.searchsorted(freq, fmin)
        freq = freq[fmin_idx:]
    return freq



import numpy as np
import pandas as pd

LATEX = {
    "idx": "ID",
    "L2_errors_VI": "L$_2$ Error",
    "len_point_CI_S11": r"$\bold{S}_{11}$ CI Length",
    "len_point_CI_re_S12": r"Re($\bold{S}_{12}$) CI Length",
    "len_point_CI_im_S12": r"Im($\bold{S}_{12}$) CI Length",
    "len_point_CI_S22": r"$\bold{S}_{22}$ CI Length",
    "coverage_pointwise": "Coverage",
}


class PSDAnalyzer:
    """
    This is a utility class to analyze the estimated PSDs, given the true PSDs.
    """

    def __init__(
            self,
            spec_true: np.ndarray,
            spectral_density_q: np.ndarray,
            idx=None,
            csv_file=None,
    ):
        """

        :param spec_true: the true psd
        :param spectral_density_q: An array containing the estimated PSD, including the lower bound (5th percentile),
        median (50th percentile), and upper bound (95th percentile) for the 90% pointwise confidence interval (CI).
        :param idx: the number of the analysis task, the default is set to 1, which is one realization
        """
        self.spec_true = spec_true
        self.spectral_density_q = spectral_density_q

        if self.spec_true.shape != spectral_density_q[0].shape:
            raise ValueError(
                "The true PSD and the estimated PSD should have the same shape. "
                f" True PSD shape: {self.spec_true.shape}, "
                f"Estimated PSD shape: {spectral_density_q[0].shape}"
            )

        self.n_freq = spectral_density_q.shape[1]
        self.idx = idx
        self.real_spec_true = self._transform_spec_true_to_real()

        # RUN ANALYSIS
        self.l2_error = self._calculate_L2_error()
        (
            self.len_point_CI_f11,
            self.len_point_CI_re_f12,
            self.len_point_CI_im_f12,
            self.len_point_CI_f22,
        ) = self._calculate_CI_length()
        self.coverage_point_CI = self._calculate_coverage()

        if csv_file:
            self.save(csv_file)

    def _transform_spec_true_to_real(self):
        real_spec_true = np.zeros_like(self.spec_true, dtype=float)
        for j in range(self.n_freq):
            real_spec_true[j] = self._complex_to_real(self.spec_true[j])
        return real_spec_true

    @staticmethod
    def _complex_to_real(matrix):
        n = matrix.shape[0]
        real_matrix = np.zeros_like(matrix, dtype=float)
        real_matrix[np.triu_indices(n)] = np.real(matrix[np.triu_indices(n)])
        real_matrix[np.tril_indices(n, -1)] = np.imag(
            matrix[np.tril_indices(n, -1)]
        )
        return real_matrix

    def _calculate_L2_error(self):
        spec_mat_median = self.spectral_density_q[1]
        N2_VI = np.empty(self.n_freq)
        for i in range(self.n_freq):
            N2_VI[i] = np.sum(
                np.diag(
                    (spec_mat_median[i, :, :] - self.spec_true[i, :, :])
                    @ (spec_mat_median[i, :, :] - self.spec_true[i, :, :])
                )
            )
        L2_VI = np.sqrt(np.mean(N2_VI))
        return L2_VI

    def _calculate_CI_length(self):
        spec_mat_upper = self.spectral_density_q[2]
        spec_mat_lower = self.spectral_density_q[0]

        len_point_CI_f11 = np.median(
            np.real(spec_mat_upper[:, 0, 0])
        ) - np.median(np.real(spec_mat_lower[:, 0, 0]))
        len_point_CI_re_f12 = np.median(
            np.real(spec_mat_upper[:, 0, 1])
        ) - np.median(np.real(spec_mat_lower[:, 0, 1]))
        len_point_CI_im_f12 = np.median(
            np.imag(spec_mat_upper[:, 1, 0])
        ) - np.median(np.imag(spec_mat_lower[:, 1, 0]))
        len_point_CI_f22 = np.median(
            np.real(spec_mat_upper[:, 1, 1])
        ) - np.median(np.real(spec_mat_lower[:, 1, 1]))

        return (
            len_point_CI_f11,
            len_point_CI_re_f12,
            len_point_CI_im_f12,
            len_point_CI_f22,
        )

    def _calculate_coverage(self):
        spec_mat_lower_real = np.zeros_like(
            self.spectral_density_q[0], dtype=float
        )
        for j in range(self.n_freq):
            spec_mat_lower_real[j] = self._complex_to_real(
                self.spectral_density_q[0][j]
            )

        spec_mat_upper_real = np.zeros_like(
            self.spectral_density_q[2], dtype=float
        )
        for j in range(self.n_freq):
            spec_mat_upper_real[j] = self._complex_to_real(
                self.spectral_density_q[2][j]
            )
        coverage_point_CI = np.mean(
            (spec_mat_lower_real <= self.real_spec_true)
            & (self.real_spec_true <= spec_mat_upper_real)
        )
        return coverage_point_CI

    def save(self, fname):
        results = {
            "idx": self.idx,
            "L2_errors_VI": self.l2_error,
            "len_point_CI_S11": self.len_point_CI_f11,
            "len_point_CI_re_S12": self.len_point_CI_re_f12,
            "len_point_CI_im_S12": self.len_point_CI_im_f12,
            "len_point_CI_S22": self.len_point_CI_f22,
            "coverage_pointwise": self.coverage_point_CI,
        }
        result_df = pd.DataFrame([results])
        result_df.to_csv(fname, index=False)
        return result_df

    def __dict__(self):
        data = {
            "L2_errors_VI": self.l2_error,
            "len_point_CI_S11": self.len_point_CI_f11,
            "len_point_CI_re_S12": self.len_point_CI_re_f12,
            "len_point_CI_im_S12": self.len_point_CI_im_f12,
            "len_point_CI_S22": self.len_point_CI_f22,
            "coverage_pointwise": self.coverage_point_CI,
        }
        if self.idx is not None:
            data["idx"] = self.idx
        return data

    @property
    def df(self):
        return pd.DataFrame([self.__dict__()])

    def _repr_html_(self):
        df = self.df.rename(columns=LATEX).T
        return df.to_html()


import itertools
from scipy.stats import median_abs_deviation
import matplotlib.pyplot as plt
import numpy as np

from ..logging import logger


def matrix_combinations(p):
    indices = range(p)
    combinations = itertools.combinations(indices, 2)
    result = [((i, j), (i, i), (j, j)) for i, j in combinations]
    return result


def uniformmax_help(sample):
    return np.abs(sample - np.median(sample)) / median_abs_deviation(sample)


def uniformmax_multi(coh_all):
    N_sample, N = coh_all.shape
    C_help = np.zeros((N_sample, N))

    for j in range(N):
        C_help[:, j] = uniformmax_help(coh_all[:, j])

    return np.max(C_help, axis=0)



def compute_coherence(pxx, pyy, pxy):
    coh_q = np.real(np.abs(pxy) ** 2) / np.real(pxx) / np.real(pyy)
    coh_uniform = None
    if len(coh_q.shape) > 1:
        coh_all = coh_q
        coh_q = np.quantile(coh_all, [0.05, 0.5, 0.95], axis=0)

        coh_median = coh_q[1]
        mad = median_abs_deviation(coh_all, axis=0, nan_policy='omit')
        mad[mad == 0] = 1e-10
        max_std_abs_dev = uniformmax_multi(coh_all)
        threshold = np.quantile(max_std_abs_dev, 0.9)
        coh_lower = coh_median - threshold * mad
        coh_upper = coh_median + threshold * mad
        coh_uniform = np.stack([coh_lower, coh_median, coh_upper], axis=0)

    return coh_q, coh_uniform





def plot_coherence(psd, freq, labels=None, ax=None, color=None, ls="-"):
    p = psd.shape[-1]
    combinations = matrix_combinations(p)
    if ax is None:
        _, ax = plt.subplots(1, 1)

    for idx, comb in enumerate(combinations):
        (i, j), (ii, _), (jj, _) = comb
        coh_q, coh_uniform = compute_coherence(
            psd[..., ii, ii], psd[..., jj, jj], psd[..., i, j]
        )
        if labels is not None:
            l = r"$C_{" + f"{labels[i]}{labels[j]}" + "}$"
        else:
            l = None
        _plot_one_coherence(
            coh_q,
            coh_uniform,
            freq,
            label=l,
            color=f"C{idx}" if color is None else color,
            ax=ax,
            ls=ls,
        )
    ax.legend()
    return ax


def _plot_one_coherence(coh, coh_uniform, freq, label, ax, color, ls="-"):
    """
    Plot the coherence between two signals.
    """
    nqt = len(coh.shape)
    if nqt > 1:
        ax.fill_between(freq, coh[0], coh[2], alpha=0.3, lw=0, color=color)
        ax.fill_between(freq, coh_uniform[0], coh_uniform[2], alpha=0.5, lw=0, color=color)
        coh_median = coh[1]
    else:
        coh_median = coh
    ax.plot(freq, coh_median, color=color, label=label, ls=ls)
    ax.set_xlabel("Frequency [Hz]")
    ax.set_ylabel(r"$C_{xy}$")
    ax.set_xlim([freq[0], freq[-1]])
    return ax



import matplotlib.pyplot as plt
import numpy as np
import rpy2.robjects as ro
import rpy2.robjects.packages as rpackages
from rpy2.robjects import r

from sgvb_psd.postproc.plot_psd import plot_psdq
from sgvb_psd.psd_estimator import PSDEstimator
from sgvb_psd.utils.sim_varma import SimVARMA

# Import the necessary R packages
vnpc_avg = rpackages.importr("vnpc.avg")
beyond_whittle = rpackages.importr("beyondWhittle")


def generate_data(seed_number: int = 0) -> np.ndarray:
    r(f"set.seed({seed_number})")
    param_ar = r["rbind"](
        ro.FloatVector([0.1, 0, 0, 0]), ro.FloatVector([0, -0.3, 0, -0.5])
    )
    param_ma = r["matrix"](nrow=2, ncol=0)
    param_sigma = r["matrix"](ro.FloatVector([1, 0.9, 0.9, 1]), nrow=2, ncol=2)
    data = beyond_whittle.sim_varma(
        model=r.list(ar=param_ar, ma=param_ma, sigma=param_sigma), n=512, d=2
    )
    return np.array(data)


def load_vnpc_results(fpath="mcmc_results.RData"):
    r["load"](fpath)
    r_obj = r["mcmc_vnp_avg"]
    result_dict = {}

    # Iterate over the ListVector items (names and values)
    for name, value in zip(r_obj.names, r_obj):
        # Convert R objects to appropriate Python/Numpy structures
        if isinstance(value, ro.vectors.FloatMatrix):
            result_dict[name] = np.array(value)
        elif isinstance(value, ro.vectors.FloatArray):
            result_dict[name] = np.array(value)
        else:
            result_dict[name] = value  # Keep as it is for other data types

    coh = np.array(
        [
            result_dict["coherence.p05"],
            result_dict["coherence.median"],
            result_dict["coherence.p95"],
        ]
    )
    data = result_dict["data"]
    psd = np.array(
        [  # pointwise
            result_dict["psd.p05"],
            result_dict["psd.median"],
            result_dict["psd.p95"],
        ]
    )

    x, d, d, nfreq = psd.shape
    # reshape to x, nfreq. d, d
    psd = np.transpose(psd, (0, 3, 1, 2))
    # cut 0th and last nfreq
    psd = psd[:, 1:-1, :, :]

    psd = psd.astype(np.complex128)
    # make lower triangle elements imaginary by multiplying by 1j
    for i in range(d):
        for j in range(d):
            if i > j:
                psd[:, :, i, j] *= 1j
    return psd, coh, data


def plot():
    pass


if __name__ == "__main__":
    mcmc_psd, mcmc_coh, data = load_vnpc_results()
    psd_estimator = PSDEstimator(
        x=data,
        N_theta=40,
        nchunks=4,
        fs=2 * np.pi,
        N_samples=1000,
        max_hyperparm_eval=1000,
    )
    psd_estimator.run(lr=0.003)
    var2 = SimVARMA(
        n_samples=int((2**9) / 4),
        var_coeffs=np.array(
            [[[0.1, 0.0], [0.0, -0.3]], [[0.0, 0], [0.0, -0.5]]]
        ),
        vma_coeffs=np.array([[[0.0, 0.0], [0.0, 0.0]]]),
        sigma=np.array([[1.0, 0.9], [0.9, 1.0]]),
    )
    axes = plot_psdq(mcmc_psd, freqs=psd_estimator.freq, color="tab:red")
    axes = psd_estimator.plot(
        off_symlog=False,
        axes=axes,
        true_psd=[var2.psd, var2.freq],
        xlims=[0, np.pi],
    )
    plt.savefig("sgvb_mcmc_simulation_compare2.png", dpi=500)



from typing import List, Tuple

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow.keras.optimizers import Adam

from ..logging import logger
from .analysis_data import AnalysisData
from .compute_psd import compute_psd

tfd = tfp.distributions
tfb = tfp.bijectors


class BayesianModel:
    def __init__(
        self,
        data: AnalysisData,
        degree_fluctuate: float = None,
        init_params: List[tf.Variable] = None,
    ):

        self.data = data

        if degree_fluctuate is None:
            degree_fluctuate = data.N_delta / 2

        # Setup hyperparams
        self.degree_fluctuate = tf.convert_to_tensor(
            degree_fluctuate, dtype=tf.float32
        )
        self.tau0 = tf.convert_to_tensor(0.01, dtype=tf.float32)
        self.c2 = tf.convert_to_tensor(4, dtype=tf.float32)
        self.sig2_alp = tf.convert_to_tensor(10, dtype=tf.float32)
        self.hyper = [self.tau0, self.c2, self.sig2_alp, self.degree_fluctuate]

        self.log_map_vals = tf.Variable(0.0)

        # convert required objects to tensors
        self.trainable_vars = self._get_trainable_vars()
        if init_params is not None:
            for i, p in enumerate(init_params):
                self.trainable_vars[i].assign(p)

        # Initialize model with MAP
        logger.debug(f"Initialized model.")

    def _get_trainable_vars(self, batch_size: int = 1) -> List[tf.Variable]:
        #
        #
        # rule:  self.trainable_vars[0, 2, 4] must be corresponding spline regression parameters for p>1
        # in 1-d case, self.trainable_vars[0] must be ga_delta parameters, no ga_theta included.

        # initial values are quite important for training

        p = int(self.data.y_ft.shape[2])
        size_delta = int(self.data.Xmat_delta.shape[1])
        size_theta = int(self.data.Xmat_theta.shape[1])

        # initializer = tf.initializers.GlorotUniform() # xavier initializer
        # initializer = tf.initializers.RandomUniform(minval=-0.5, maxval=0.5)
        # initializer = tf.initializers.zeros()

        # better to have deterministic inital on reg coef to control
        ga_initializer = tf.initializers.zeros()
        ga_initializer_para = tf.initializers.constant(value=0.0)
        ga_initializer_para2 = tf.initializers.constant(value=0.0)
        if size_delta <= 10:
            cvec_d = 0.0
        else:
            cvec_d = tf.concat(
                [tf.zeros(10 - 2) + 0.0, tf.zeros(size_delta - 10) + 1.0], 0
            )
        if size_theta <= 10:
            cvec_o = 0.5
        else:
            cvec_o = tf.concat(
                [tf.zeros(10) + 0.5, tf.zeros(size_theta - 10) + 1.5], 0
            )

        ga_delta = tf.Variable(
            ga_initializer_para(
                shape=(batch_size, p, size_delta), dtype=tf.float32
            ),
            name="ga_delta",
            trainable=True,
            dtype=tf.float32,
        )
        lla_delta = tf.Variable(
            ga_initializer(
                shape=(batch_size, p, size_theta - 2), dtype=tf.float32
            )
            - cvec_d,
            name="lla_delta",
            trainable=True,
            dtype=tf.float32,
        )
        ltau = tf.Variable(
            ga_initializer(shape=(batch_size, p, 1), dtype=tf.float32) - 1,
            name="ltau",
            trainable=True,
            dtype=tf.float32,
        )

        nn = int(p * (p - 1) / 2)  # number of thetas in the model
        ga_theta_re = tf.Variable(
            ga_initializer_para2(
                shape=(batch_size, nn, size_theta), dtype=tf.float32
            ),
            name="ga_theta_re",
            trainable=True,
            dtype=tf.float32,
        )
        ga_theta_im = tf.Variable(
            ga_initializer_para2(
                shape=(batch_size, nn, size_theta), dtype=tf.float32
            ),
            name="ga_theta_im",
            trainable=True,
            dtype=tf.float32,
        )

        lla_theta_re = tf.Variable(
            ga_initializer(
                shape=(batch_size, nn, size_theta), dtype=tf.float32
            )
            - cvec_o,
            name="lla_theta_re",
            trainable=True,
            dtype=tf.float32,
        )
        lla_theta_im = tf.Variable(
            ga_initializer(
                shape=(batch_size, nn, size_theta), dtype=tf.float32
            )
            - cvec_o,
            name="lla_theta_im",
            trainable=True,
            dtype=tf.float32,
        )

        ltau_theta = tf.Variable(
            ga_initializer(shape=(batch_size, nn, 1), dtype=tf.float32) - 1.5,
            name="ltau_theta",
            trainable=True,
            dtype=tf.float32,
        )

        # params:          self.trainable_vars (ga_delta, lla_delta,
        #                                       ga_theta_re, lla_theta_re,
        #                                       ga_theta_im, lla_theta_im,
        #                                       ltau, ltau_theta)
        return [
            ga_delta,
            lla_delta,
            ga_theta_re,
            lla_theta_re,
            ga_theta_im,
            lla_theta_im,
            ltau,
            ltau_theta,
        ]

    def loglik(self, params: List[tf.Variable]) -> tf.float32:
        # y_re:            self.y_re
        # y_im:            self.y_im
        # Z_:              self.Zar
        # X_:              self.Xmat
        # params:          self.trainable_vars (ga_delta, xxx,
        #                                       ga_theta_re, xxx,
        #                                       ga_theta_im, xxx, ...)
        # each of params is a 3-d tensor with sample_size as the fist dim.
        # self.trainable_vars[:,[0, 2, 4]] must be corresponding spline regression parameters

        xγ = tf.matmul(
            self.data.Xmat_delta, tf.transpose(params[0], [0, 2, 1])
        )
        sum_xγ = -tf.reduce_sum(xγ, [1, 2])
        exp_xγ_inv = tf.exp(-xγ)

        xα = tf.matmul(
            self.data.Xmat_theta, tf.transpose(params[2], [0, 2, 1])
        )  # no need \ here
        xβ = tf.matmul(
            self.data.Xmat_theta, tf.transpose(params[4], [0, 2, 1])
        )

        # Z = Sum [(xα + i xβ) * y]
        Z_theta_re = tf.linalg.matvec(
            tf.expand_dims(self.data.Z_re, 0), xα
        ) - tf.linalg.matvec(tf.expand_dims(self.data.Z_im, 0), xβ)
        Z_theta_im = tf.linalg.matvec(
            tf.expand_dims(self.data.Z_re, 0), xβ
        ) + tf.linalg.matvec(tf.expand_dims(self.data.Z_im, 0), xα)

        u_re = self.data.y_re - Z_theta_re
        u_im = self.data.y_im - Z_theta_im

        numerator = tf.square(u_re) + tf.square(u_im)
        internal = tf.multiply(numerator, exp_xγ_inv)
        tmp2_ = -tf.reduce_sum(internal, [-2, -1])  # sum over p and freq
        log_lik = tf.reduce_sum(sum_xγ + tmp2_)  # sum over all LnL
        return log_lik

    def logpost(self, params: List[tf.Variable]) -> tf.float32:
        return self.loglik(params) + self.logprior(params)

    #
    # Model training one step
    #
    def map_train_step(
        self, optimizer: Adam
    ) -> tf.float32:  # one training step to get close to MAP
        with tf.GradientTape() as tape:
            self.log_map_vals = -1 * self.logpost(self.trainable_vars)

        grads = tape.gradient(self.log_map_vals, self.trainable_vars)
        grads_and_vars = [
            (g, v) for g, v in zip(grads, self.trainable_vars) if g is not None
        ]
        if grads_and_vars:
            optimizer.apply_gradients(grads_and_vars)

        self.log_map_vals *= -1  # return POSITIVE log posterior
        return self.log_map_vals

    # For new prior strategy, need new createModelVariables() and logprior()

    def logprior(self, params):
        # hyper:           list of hyperparameters (tau0, c2, sig2_alp, degree_fluctuate)
        # params:          self.trainable_vars (ga_delta, lla_delta,
        #                                       ga_theta_re, lla_theta_re,
        #                                       ga_theta_im, lla_theta_im,
        #                                       ltau, ltau_theta)
        # each of params is a 3-d tensor with sample_size as the fist dim.
        # self.trainable_vars[:,[0, 2, 4]] must be corresponding spline regression parameters
        Sigma1 = tf.multiply(
            tf.eye(tf.constant(2), dtype=tf.float32), self.hyper[2]
        )
        priorDist1 = tfd.MultivariateNormalTriL(
            scale_tril=tf.linalg.cholesky(Sigma1)
        )  # can also use tfd.MultivariateNormalDiag

        Sigm = tfb.Sigmoid()
        s_la_alp = Sigm(
            -tf.range(1, params[1].shape[-1] + 1.0, dtype=tf.float32)
            + self.hyper[3]
        )
        priorDist_la_alp = tfd.HalfCauchy(tf.constant(0, tf.float32), s_la_alp)

        s_la_theta = Sigm(
            -tf.range(1, params[3].shape[-1] + 1.0, dtype=tf.float32)
            + self.hyper[3]
        )
        priorDist_la_theta = tfd.HalfCauchy(
            tf.constant(0, tf.float32), s_la_theta
        )

        a2 = tf.square(tf.exp(params[1]))
        Sigma2i_diag = tf.divide(
            tf.multiply(
                tf.multiply(a2, tf.square(tf.exp(params[6]))), self.hyper[1]
            ),
            tf.multiply(a2, tf.square(tf.exp(params[6]))) + self.hyper[1],
        )

        priorDist2 = tfd.MultivariateNormalDiag(scale_diag=Sigma2i_diag)

        lpriorAlp_delt = tf.reduce_sum(
            priorDist1.log_prob(params[0][:, :, 0:2]), [1]
        )  #
        lprior_delt = tf.reduce_sum(
            priorDist2.log_prob(params[0][:, :, 2:]), [1]
        )  # only 2 dim due to log_prob rm the event_shape dim
        lpriorla_delt = tf.reduce_sum(
            priorDist_la_alp.log_prob(tf.exp(params[1])), [1, 2]
        ) + tf.reduce_sum(params[1], [1, 2])
        lpriorDel = lprior_delt + lpriorla_delt + lpriorAlp_delt

        a3 = tf.square(tf.exp(params[3]))
        Sigma3i_diag = tf.divide(
            tf.multiply(
                tf.multiply(a3, tf.square(tf.exp(params[7]))), self.hyper[1]
            ),
            tf.multiply(a3, tf.square(tf.exp(params[7]))) + self.hyper[1],
        )

        priorDist3 = tfd.MultivariateNormalDiag(scale_diag=Sigma3i_diag)

        lprior_thet_re = tf.reduce_sum(priorDist3.log_prob(params[2]), [1])
        lpriorla_thet_re = tf.reduce_sum(
            priorDist_la_theta.log_prob(tf.exp(params[3])), [1, 2]
        ) + tf.reduce_sum(params[3], [1, 2])
        lpriorThe_re = lprior_thet_re + lpriorla_thet_re

        a4 = tf.square(tf.exp(params[5]))
        Sigma4i_diag = tf.divide(
            tf.multiply(
                tf.multiply(a4, tf.square(tf.exp(params[7]))), self.hyper[1]
            ),
            tf.multiply(a4, tf.square(tf.exp(params[7]))) + self.hyper[1],
        )

        priorDist4 = tfd.MultivariateNormalDiag(scale_diag=Sigma4i_diag)

        lprior_thet_im = tf.reduce_sum(priorDist4.log_prob(params[4]), [1])
        lpriorla_thet_im = tf.reduce_sum(
            priorDist_la_theta.log_prob(tf.exp(params[5])), [1, 2]
        ) + tf.reduce_sum(params[5], [1, 2])
        lpriorThe_im = lprior_thet_im + lpriorla_thet_im

        priorDist_tau = tfd.HalfCauchy(
            tf.constant(0, tf.float32), self.hyper[0]
        )
        logPrior = (
            lpriorDel
            + lpriorThe_re
            + lpriorThe_im
            + tf.reduce_sum(
                priorDist_tau.log_prob(tf.exp(params[6])) + params[6], [1, 2]
            )
            + tf.reduce_sum(
                priorDist_tau.log_prob(tf.exp(params[7])) + params[7], [1, 2]
            )
        )
        return logPrior

    def compute_psd(
        self,
        vi_samples: List[tf.Tensor],
        quantiles=[0.05, 0.5, 0.95],
        psd_scaling=1.0,
        fs=None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return compute_psd(
            self.data.Xmat_delta,
            self.data.Xmat_theta,
            self.data.p,
            vi_samples,
            quantiles,
            psd_scaling,
            fs,
        )



"""
Run a simulation study to compare the performance of the PSD estimator

(currently uses lower settings than in the paper for faster execution).
"""

from sgvb_psd.psd_estimator import PSDEstimator
from sgvb_psd.postproc import plot_psdq
from sgvb_psd.postproc.psd_analyzer import PSDAnalyzer
from sgvb_psd.utils.sim_varma import SimVARMA
import numpy as np
from tqdm.auto import tqdm
import matplotlib.pyplot as plt
import pandas as pd
from sgvb_psd.logging import logger
import os

N_EXPERIMENTS = 10

SIM_KWGS = dict(
    sigma=np.array([[1.0, 0.9], [0.9, 1.0]]),
    var_coeffs=np.array([[[0.5, 0.0], [0.0, -0.3]], [[0.0, 0.0], [0.0, -0.5]]]),
    vma_coeffs=np.array([[[1.0, 0.0], [0.0, 1.0]]]),
    n_samples=1024,
)
VI_KWGS = dict(
    N_theta=35,
    nchunks=1,
    ntrain_map=1000,
    max_hyperparm_eval=1,
    fs=2 * np.pi,
    seed=0,
)

logger.setLevel('INFO')
OUTDIR = 'out_simulation_results'
os.makedirs(OUTDIR, exist_ok=True)

def run_simulation(idx):
    sim = SimVARMA(**SIM_KWGS, seed=idx)
    optim = PSDEstimator(**VI_KWGS, x=sim.data)
    optim.run(lr=0.003)
    plot_simulation_psd(
        f'{OUTDIR}/simulation_{idx}.png',
        optim,
        sim,
    )
    summary = PSDAnalyzer(
        sim.psd,
        optim.pointwise_ci,
        idx=idx,
    ).__dict__()
    summary['time'] = optim.runtimes['train']
    return {k: summary[k] for k in ['idx', 'L2_errors_VI', 'time']}

def plot_simulation_psd(fname, optim:PSDEstimator, sim:SimVARMA):
    fig = optim.plot(
        true_psd=[sim.psd, sim.freq],
        off_symlog=False,
        xlims=[0, np.pi],
        quantiles='pointwise',
    )
    plot_psdq(
        optim.uniform_ci,
        freqs=optim.freq,
        axs=fig,
        color='red',
        ls='--',
    )
    plt.savefig(fname)



def run_all_simulations(n_experiments=N_EXPERIMENTS):
    results = []
    for idx in tqdm(range(n_experiments)):
        results.append(run_simulation(idx))
    return pd.DataFrame(results)


def plot_results(results: pd.DataFrame):
    """Ploot violin plots of L2 errors and training times."""
    fig, ax = plt.subplots(1, 2, figsize=(12, 6))
    results.boxplot('L2_errors_VI', ax=ax[0])
    results.boxplot('time', ax=ax[1])
    ax[0].set_ylabel('L2 error')
    ax[1].set_ylabel('time')
    # add a title
    fig.suptitle(f"{N_EXPERIMENTS} simulations")
    return fig


def main():
    results = run_all_simulations()
    plot_results(results)
    plt.savefig('simulation_results.png')


if __name__ == '__main__':
    main()



import click


@click.command("sgvb_benchmark")
@click.option(
    '-s',
    '--minlog2n',
    default=3,
    help='Minimum log2(n) for the benchmark.',
)
@click.option(
    '-e',
    '--maxlog2n',
    default=4,
    help='Maximum log2(n) for the benchmark.',
)
@click.option(
    '-f',
    '--fname',
    default='timings.txt',
    help='Filename for the benchmark results.',
)
@click.option(
    '-n',
    '--nrep',
    default=5,
    help='Number of repetitions for the benchmark.',
)
def cli(minlog2n, maxlog2n, fname, nrep):
    """
    Command line interface for benchmarking the PSD estimator.
    """
    from .benchmark import benchmark

    benchmark(minLog2n=minlog2n, maxLog2n=maxlog2n, fname=fname, nrep=nrep)



import matplotlib.pyplot as plt
import numpy as np

N = 10000
N_CHANNELS = 2
n_blocks = 6
n_per_block = N // n_blocks

t = np.linspace(0, 100, N)
x = np.random.random((N_CHANNELS, n_blocks, n_per_block))

fig, axs = plt.subplots(N_CHANNELS, 1, figsize=(3, 2))


for i in range(N_CHANNELS):
    for j in range(n_blocks):
        _t = t[j * n_per_block : (j + 1) * n_per_block]
        axs[i].plot(_t, np.sin(25 * _t) + x[i, j], alpha=1)
    axs[i].annotate(
        f"Channel {i+1}",
        xy=(0.1, 0.7),
        xycoords="axes fraction",
        # fontcolor white
        color="black",
    )
    axs[i].set_yticks([])
    axs[i].set_xticks([])

fig.legend(
    handles=[
        plt.Line2D([0], [0], color=f"C{i}", lw=3, label=f"Block {i+1}")
        for i in range(n_blocks)
    ],
    loc="outside center right",
    frameon=False,
)
axs[-1].set_xlabel("Time [s]")

plt.subplots_adjust(hspace=0, right=0.5)
plt.savefig("./out_plots/plot_block.png")



import time
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from hyperopt import fmin, hp, tpe

from .backend import BayesianModel, ViRunner
from .logging import logger
from .postproc import plot_coherence
from .postproc.plot_losses import plot_losses
from .postproc.plot_psd import plot_psd
from .utils.periodogram import get_periodogram
from .utils.tf_utils import set_seed
from .utils.utils import get_freq


class PSDEstimator:
    """
    A class for estimating the posterior Power Spectral Density (PSD) using Stochastic Gradient Variational Bayes (SGVB).

    This class implements a two-step process:
    1. Optimize the learning rate to maximize the posterior and Evidence Lower Bound (ELBO).
    2. Use the optimized learning rate to estimate the posterior PSD.

    The main interface is the run() method, which returns the posterior PSD and its quantiles.

    :ivar N_theta: Number of basis functions for the theta component.
    :vartype N_theta: int
    :ivar N_samples: Number of parameters sampled from the surrogate distribution.
    :vartype N_samples: int
    :ivar nchunks: Number of blocks the multivariate time series is divided into.
    :vartype nchunks: int
    :ivar ntrain_map: Number of iterations in gradient ascent for Maximum A Posteriori (MAP) estimation.
    :vartype ntrain_map: int
    :ivar fs: Sampling frequency of the input data.
    :vartype fs: float
    :ivar lr_range: Range of learning rates to consider during optimization.
    :vartype lr_range: tuple
    :ivar psd_scaling: Scaling factor for the input data.
    :vartype psd_scaling: numpy.ndarray
    :ivar psd_offset: Offset for the input data.
    :vartype psd_offset: numpy.ndarray
    :ivar x: Normalized input multivariate time series.
    :vartype x: numpy.ndarray
    :ivar n: Number of time points in the input data.
    :vartype n: int
    :ivar p: Number of variables in the multivariate time series.
    :vartype p: int
    :ivar fmax_for_analysis: Maximum frequency in the frequency domain to be analyzed.
    :vartype fmax_for_analysis: int
    :ivar pdgrm: Periodogram of the input data.
    :vartype pdgrm: numpy.ndarray
    :ivar pdgrm_freq: Frequencies corresponding to the periodogram.
    :vartype pdgrm_freq: numpy.ndarray
    :ivar max_hyperparm_eval: Number of evaluations in hyperparameter optimization.
    :vartype max_hyperparm_eval: int
    :ivar degree_fluctuate: Hyperparameter from the prior, used when dealing with a large number of basis functions.
    :vartype degree_fluctuate: float
    :ivar model: Trained model object.
    :vartype model: object
    :ivar samps: Samples drawn from the posterior distribution.
    :vartype samps: numpy.ndarray
    :ivar vi_losses: Variational Inference losses during training.
    :vartype vi_losses: numpy.ndarray
    :ivar psd_quantiles: Quantiles of the estimated PSD.
    :vartype pointwise_ci: numpy.ndarray
    :ivar psd_all: All estimated PSDs.
    :vartype psd_all: numpy.ndarray
    :ivar inference_runner: Object for running the variational inference.
    :vartype inference_runner: ViRunner
    :ivar optimal_lr: Optimized learning rate.
    :vartype optimal_lr: float
    :ivar runtimes: Runtime of the different steps in the estimation process.
    :vartype runtimes: dict
    :ivar n_elbo_maximisation_steps: Number of steps for maximising the ELBO.
    :vartype n_elbo_maximisation_steps: int
    """

    def __init__(
        self,
        x,
        N_theta=30,
        nchunks=1,
        ntrain_map=10000,
        N_samples=500,
        fs=1.0,
        max_hyperparm_eval=100,
        fmax_for_analysis=None,
        degree_fluctuate=None,
        seed=None,
        lr_range=(0.002, 0.02),
        n_elbo_maximisation_steps=500,
        init_params=None,
        fmin_for_analysis=None,
    ):
        """
        Initialize the PSDEstimator.

        :param x: Input multivariate time series of shape: (nsamp, p).
        :type x: numpy.ndarray
        :param N_theta: Number of basis functions for the theta component, defaults to 30.
        :type N_theta: int, optional
        :param nchunks: Number of blocks to divide the multivariate time series into, defaults to 1.
        :type nchunks: int, optional
        :param ntrain_map: Number of iterations in gradient ascent for MAP, defaults to 10000.
        :type ntrain_map: int, optional
        :param N_samples: Number of parameters sampled from the surrogate distribution, defaults to 500.
        :type N_samples: int, optional
        :param fs: Sampling frequency, defaults to 1.0.
        :type fs: float, optional
        :param max_hyperparm_eval: Number of evaluations in hyperparameter optimization, defaults to 100.
        :type max_hyperparm_eval: int, optional
        :param fmax_for_analysis: Maximum frequency to analyze in the frequency domain, defaults to None.
        :type fmax_for_analysis: int, optional
        :param degree_fluctuate: Hyperparameter from the prior, defaults to None.
        :type degree_fluctuate: float, optional
        :param seed: Random seed for reproducibility, defaults to None.
        :type seed: int, optional
        :param lr_range: Range of learning rates to consider during optimization, defaults to (0.002, 0.02).
        :type lr_range: tuple, optional
        :param n_elbo_maximisation_steps: Number of steps for maximizing the ELBO, defaults to 1000.
        :type n_elbo_maximisation_steps: int, optional
        """

        if seed is not None:
            set_seed(seed)

        self.N_theta = N_theta
        self.N_samples = N_samples
        self.nchunks = nchunks
        self.ntrain_map = ntrain_map
        self.n_elbo_maximisation_steps = n_elbo_maximisation_steps

        self.fs = fs
        self.lr_range = lr_range

        # normalize the data
        self.psd_scaling = np.std(x)
        self.psd_offset = np.mean(x)
        self.x = (x - self.psd_offset) / self.psd_scaling
        self.n, self.p = x.shape
        if fmax_for_analysis is None:
            fmax_for_analysis = self.n // 2
        self.fmax_for_analysis = fmax_for_analysis
        self.fmin_for_analysis = fmin_for_analysis

        self.pdgrm, self.pdgrm_freq = get_periodogram(self.x, fs=self.fs)
        self.pdgrm = self.pdgrm * self.psd_scaling**2
        self.max_hyperparm_eval = max_hyperparm_eval
        self.degree_fluctuate = degree_fluctuate

        if self.nchunks > 1:
            logger.info(
                f"Dividing data {self.x.shape} into "
                f"({self.nchunks}, {self.nt_per_chunk}, {self.p}) chunks"
            )
            # if nchunks is not a power of 2, wil be slower
            if np.log2(nchunks) % 1 != 0:
                logger.warning("nchunks must be a power of 2 for faster FFTs")

        if self.fmax_for_analysis < self.nt_per_chunk // 2:
            logger.info(
                f"Reducing the number of frequencies to be analyzed from "
                f"{self.nt_per_chunk // 2} to {self.fmax_for_analysis}..."
            )

        logger.info(
            f"Final PSD will be of shape: {self.nfreq_per_chunk} x {self.p} x {self.p}"
        )

        # Internal variables
        self.model: BayesianModel = None
        self.samps = None
        self.pointwise_ci = None
        self.uniform_ci = None
        self.psd_all = None
        self.runtimes = {}
        self.inference_runner = ViRunner(
            self.x,
            N_theta=self.N_theta,
            nchunks=self.nchunks,
            fmax_for_analysis=self.fmax_for_analysis,
            fmin_for_analysis=self.fmin_for_analysis,
            degree_fluctuate=self.degree_fluctuate,
            fs=self.fs,
            init_params=init_params,
        )

    def _learning_rate_optimisation_objective(self, lr):
        """
        Objective function for hyperparameter optimization of the learning rate for MAP.

        Args:
            lr (dict): Dictionary containing the learning rate to be optimized.

        Returns:
            float: ELBO log_map_vals.
        """
        vi_losses, _, _, _ = self.inference_runner.run(
            lr_map=lr["lr_map"],
            ntrain_map=self.ntrain_map,
            inference_size=self.N_samples,
            n_elbo_maximisation_steps=self.n_elbo_maximisation_steps,
        )
        return vi_losses[-1].numpy()

    def _find_optimal_learing_rate(self):
        """
        Find the optimal learning rate using hyperopt.

        This method uses the TPE algorithm to optimize the learning rate.
        """
        self.optimal_lr = fmin(
            self._learning_rate_optimisation_objective,
            space={"lr_map": hp.uniform("lr_map", *self.lr_range)},
            algo=tpe.suggest,
            max_evals=self.max_hyperparm_eval,
        )["lr_map"]
        logger.info(f"Optimal learning rate: {self.optimal_lr:.4e}")

    def _train_model(self):
        """
        Train the model using the optimal learning rate.

        This method runs the variational inference to estimate the posterior PSD.
        """
        _, _, self.model, self.samps = self.inference_runner.run(
            lr_map=self.optimal_lr,
            ntrain_map=self.ntrain_map,
            inference_size=self.N_samples,
            n_elbo_maximisation_steps=self.n_elbo_maximisation_steps,
        )

    def run(self, lr=None) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Run the SGVB algorithm to estimate the posterior PSD.

        This method either uses a provided learning rate or finds the optimal one,
        then trains the model and computes the posterior PSD.

        :param lr: Learning rate for MAP. If None, optimal rate is found, defaults to None.
        :type lr: float, optional
        :return: Tuple containing the posterior PSD and pointwise, and uniform quantiles of the PSD.
        :rtype: tuple(numpy.ndarray, numpy.ndarray)
        """
        times = {}

        if lr:
            logger.info(f"Using provided learning rate: {lr}")
            self.optimal_lr = lr
        else:
            logger.info("Running hyperopt to find optimal learning rate")
            t0 = time.time()
            self._find_optimal_learing_rate()
            times["lr"] = time.time() - t0
            logger.info(f"Optimal learning rate found in {times['lr']:.2f}s")

        logger.info("Training model")
        t0 = time.time()
        self._train_model()
        times["train"] = time.time() - t0
        logger.info(f"Model trained in {times['train']:.2f}s")

        logger.info("Computing posterior PSDs")
        (
            self.psd_all,
            self.pointwise_ci,
            self.uniform_ci,
        ) = self.model.compute_psd(
            self.samps, psd_scaling=self.psd_scaling, fs=self.fs
        )
        self.runtimes = times
        return self.psd_all, self.pointwise_ci, self.uniform_ci

    @property
    def freq(self) -> np.ndarray:
        """
        Get the frequencies per chunk of the PSD estimate.

        Returns:
            np.ndarray: Array of frequencies.
        """
        if not hasattr(self, "_freq"):
            self._freq = get_freq(
                fs=self.fs,
                n_time_samples=self.nt_per_chunk,
                fmax=self.fmax_for_analysis,
                fmin=self.fmin_for_analysis,
            )
        return self._freq

    @property
    def nt_per_chunk(self) -> int:
        """Return the number of time-points per chunk"""
        return self.n // self.nchunks

    @property
    def nfreq_per_chunk(self) -> int:
        """Return the number of frequencies per chunk"""
        return len(self.freq)

    def plot(
        self,
        true_psd=None,
        quantiles="pointwise",
        plot_periodogram=True,
        tick_ln=5,
        diag_spline_thickness=2,
        xlims=None,
        diag_ylims=None,
        off_ylims=None,
        diag_log=True,
        off_symlog=True,
        sylmog_thresh=1e-49,
        **kwargs,
    ) -> np.ndarray[plt.Axes]:
        """
        Plot the estimated PSD, periodogram, and true PSD (if provided).

        :param true_psd: True PSD and freq to plot for comparison (true_psd, true_freq)
        :type true_psd: tuple, optional
        :param quantiles: Type of quantiles ('pointwise', 'uniform') to plot, defaults to 'pointwise'
        :type quantiles: str, optional
        :param plot_periodogram: Whether to plot the periodogram
        :type plot_periodogram: bool
        :param tick_ln: Length of the ticks, defaults to 5
        :type tick_ln: int, optional
        :param diag_spline_thickness: Thickness of the diagonal spline, defaults to 2
        :type diag_spline_thickness: int, optional
        :param xlims: Limits for the x-axis
        :type xlims: tuple, optional
        :param diag_ylims: Limits for the diagonal
        :type diag_ylims: tuple, optional
        :param off_ylims: Limits for the off-diagonal
        :type off_ylims: tuple, optional
        :param diag_log: Whether to use a log scale for the diagonal, defaults to True
        :type diag_log: bool, optional
        :param off_symlog: Whether to use a symlog scale for the off-diagonal, defaults to True
        :type off_symlog: bool, optional
        :param sylmog_thresh: Threshold for symlog, defaults to 1e-49
        :type sylmog_thresh: float, optional
        :return: Matplotlib Axes object
        :rtype: numpy.ndarray
        """
        all_kwargs = dict(
            tick_ln=tick_ln,
            diag_spline_thickness=diag_spline_thickness,
            xlims=xlims,
            diag_ylims=diag_ylims,
            off_ylims=off_ylims,
            diag_log=diag_log,
            off_symlog=off_symlog,
            sylmog_thresh=sylmog_thresh,
            **kwargs,
        )
        pdgrm = [self.pdgrm, self.pdgrm_freq] if plot_periodogram else None

        ci = self.pointwise_ci
        if quantiles == "uniform":
            ci = self.uniform_ci

        return plot_psd(
            psdq=[ci, self.freq],
            pdgrm=pdgrm,
            true_psd=true_psd,
            **all_kwargs,
        )

    def plot_coherence(self, true_psd=None, **kwargs) -> np.ndarray[plt.Axes]:
        """
        Plot the coherence of the estimated PSD.

        :param true_psd: True PSD to plot for comparison
        :type true_psd: tuple, optional
        :param kwargs: Additional keyword arguments for plotting
        :return: Matplotlib Axes object
        :rtype: numpy.ndarray[plt.Axes]
        """
        labels = kwargs.pop("labels", "123456789")
        ax = plot_coherence(self.psd_all, self.freq, **kwargs, labels=labels)
        if true_psd is not None:
            ax = plot_coherence(
                true_psd[0], true_psd[1], **kwargs, ax=ax, ls="--", color="k"
            )
        return ax

    def plot_vi_losses(self) -> plt.Axes:
        """
        Plot the variational inference losses.

        :return: Matplotlib Axes object
        :rtype: plt.Axes
        """
        return plot_losses(
            map_losses=self.inference_runner.kdl_losses,
            kdl_losses=self.inference_runner.lp,
            map_timing=self.inference_runner.map_time,
            kdl_timing=self.inference_runner.vi_time,
        )

    def sample_posterior(self, n_samples: int = 1000) -> np.ndarray:
        """
        Sample the posterior distribution.

        :param n_samples: Number of samples to draw, defaults to 1000
        :type n_samples: int, optional
        :return: Samples from the posterior distribution
        :rtype: np.ndarray
        """
        spline_params = self.inference_runner.surrogate_posterior.sample(
            n_samples
        )
        psd = self.model.compute_psd(
            spline_params, psd_scaling=self.psd_scaling, fs=self.fs
        )
        return spline_params, psd



import os
import random

import numpy as np
import tensorflow as tf
from ..logging import logger


def set_seed(seed: int):
    os.environ["PYTHONHASHSEED"] = str(seed)
    tf.random.set_seed(seed)
    tf.compat.v2.random.set_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    logger.debug(f"Seed set to {seed}")



import matplotlib.pyplot as plt
import numpy as np

from ..postproc.plot_psd import format_axes, plot_peridogram, plot_single_psd
from ..utils.periodogram import get_periodogram, get_welch_periodogram


class SimVARMA:
    """
    Simulate Vector Autoregressive Moving Average (VARMA) processes and compute related spectral properties.
    """

    def __init__(
        self,
        n_samples: int,
        var_coeffs: np.ndarray,
        vma_coeffs: np.ndarray,
        sigma=np.array([1.0]),
        seed=None,
    ):
        """
        Initialize the SimVARMA class.

        Args:
            n_samples (int): Number of samples to generate.
            var_coeffs (np.ndarray): VAR coefficient array.
            vma_coeffs (np.ndarray): VMA coefficient array.
            sigma (np.ndarray): Covariance matrix or scalar variance.
        """
        self.n_samples = n_samples
        self.var_coeffs = var_coeffs
        self.vma_coeffs = vma_coeffs
        self.sigma = sigma
        self.dim = vma_coeffs.shape[1]
        self.psd_scaling = 1
        self.n_freq_samples = n_samples // 2

        self.fs = 2 * np.pi
        self.freq = (
            np.linspace(0, 0.5, self.n_freq_samples, endpoint=False)[1:] * self.fs
        )
        self.data = None  # set in "resimulate"
        self.periodogram = None  # set in "resimulate"
        self.welch_psd = None  # set in "resimulate"
        self.welch_f = None  # set in "resimulate"
        self.resimulate(seed=seed)

        self.psd = _calculate_true_varma_psd(
            self.n_freq_samples,
            self.dim,
            self.var_coeffs,
            self.vma_coeffs,
            self.sigma,
        )

    def resimulate(self, seed=None):
        """
        Simulate VARMA process.

        Args:
            seed (int, optional): Random seed for reproducibility.

        Returns:
            np.ndarray: Simulated VARMA process data.
        """
        if seed is not None:
            np.random.seed(seed)

        lag_ma = self.vma_coeffs.shape[0]
        lag_ar = self.var_coeffs.shape[0]

        if self.sigma.shape[0] == 1:
            cov_matrix = np.identity(self.dim) * self.sigma
        else:
            cov_matrix = self.sigma

        x_init = np.zeros((lag_ar + 1, self.dim))
        x = np.empty((self.n_samples + 101, self.dim))
        x[:] = np.nan
        x[: lag_ar + 1] = x_init
        epsilon = np.random.multivariate_normal(
            np.zeros(self.dim), cov_matrix, size=[lag_ma]
        )

        for i in range(lag_ar + 1, x.shape[0]):
            epsilon = np.concatenate(
                [
                    np.random.multivariate_normal(
                        np.zeros(self.dim), cov_matrix, size=[1]
                    ),
                    epsilon[:-1],
                ]
            )
            x[i] = np.sum(
                np.matmul(
                    self.var_coeffs,
                    x[i - 1 : i - lag_ar - 1 : -1][..., np.newaxis],
                ),
                axis=(0, -1),
            ) + np.sum(
                np.matmul(self.vma_coeffs, epsilon[..., np.newaxis]),
                axis=(0, -1),
            )

        self.data = x[101:]
        self.periodogram, self.pdgrm_freq = get_periodogram(
            self.data, fs=self.fs, psd_scaling=self.psd_scaling
        )
        self.compute_welch_periodogram(n_chunks=1)

    def compute_welch_periodogram(self, n_chunks=1):
        (self.welch_psd, self.welch_f) = get_welch_periodogram(
            self.data, fs=self.fs, n_chunks=n_chunks
        )

    def plot(self, axs=None, welch_nchunks=None, **kwargs):
        kwargs["off_symlog"] = kwargs.get("off_symlog", False)
        axs = plot_peridogram(self.periodogram, self.pdgrm_freq, axs=axs, **kwargs)
        if welch_nchunks is not None:
            self.compute_welch_periodogram(n_chunks=welch_nchunks)
            axs = plot_peridogram(
                self.welch_psd,
                self.welch_f,
                axs=axs,
                **kwargs,
                color="gray",
                alpha=0.5,
                zorder=-1,
            )
        plot_single_psd(self.psd, self.freq, axs=axs, **kwargs)
        format_axes(axs, **kwargs)
        return axs

    def _repr_html_(self):
        """
        Return an HTML representation of the VARMA process for rendering in notebooks and Sphinx documentation.
        """
        p = self.var_coeffs.shape[0]  # VAR order
        q = self.vma_coeffs.shape[0]  # VMA order

        html = f"""
        <div style="font-family: Arial, sans-serif;">
            <h3>VARMA({p}, {q}) Process</h3>
            <p>X<sub>t</sub> = """

        # VAR part
        if p > 0:
            for i in range(p):
                html += f"Φ<sub>{i + 1}</sub>X<sub>t-{i + 1}</sub> + "

        # VMA part
        html += "ε<sub>t</sub> + "
        if q > 0:
            for i in range(q):
                html += f"Θ<sub>{i + 1}</sub>ε<sub>t-{i + 1}</sub>"
                if i < q - 1:
                    html += " + "

        html += "</p>"
        html += "<p>ε<sub>t</sub> ~ N(0, Σ)</p>"

        # VAR coefficients
        html += "<h4>VAR coefficients:</h4>"
        for i in range(p):
            html += f"<p>Φ<sub>{i + 1}</sub> = {self._matrix_to_html(self.var_coeffs[i])}</p>"

        # VMA coefficients
        html += "<h4>VMA coefficients:</h4>"
        for i in range(q):
            html += f"<p>Θ<sub>{i + 1}</sub> = {self._matrix_to_html(self.vma_coeffs[i])}</p>"

        # Sigma
        html += "<h4>Covariance matrix:</h4>"
        html += f"<p>Σ = {self._matrix_to_html(self.sigma)}</p>"

        html += "</div>"
        return html

    def _matrix_to_html(self, matrix):
        """
        Convert a numpy array to an HTML table representation.
        """
        html = "<table style='border-collapse: collapse;'>"
        for row in matrix:
            html += "<tr>"
            for val in row:
                html += f"<td style='border: 1px solid black; padding: 4px;'>{val:.2f}</td>"
            html += "</tr>"
        html += "</table>"
        return html


def _calculate_true_varma_psd(
    n_samples: int,
    dim: int,
    var_coeffs: np.ndarray,
    vma_coeffs: np.ndarray,
    sigma: np.ndarray,
) -> np.ndarray:
    """
    Calculate the spectral matrix for given frequencies.

    Args:
        n_samples int: Number of samples to generate for the true PSD (up to 0.5).
        var_coeffs (np.ndarray): VAR coefficient array.
        vma_coeffs (np.ndarray): VMA coefficient array.
        sigma (np.ndarray): Covariance matrix or scalar variance.

    Returns:
        np.ndarray: VARMA spectral matrix (PSD) for freq from 0 to 0.5.
    """
    freq = np.linspace(0, 0.5, n_samples, endpoint=False)[1:]
    spec_matrix = np.apply_along_axis(
        lambda f: _calculate_spec_matrix_helper(
            f, dim, var_coeffs, vma_coeffs, sigma
        ),
        axis=1,
        arr=freq.reshape(-1, 1),
    )
    return spec_matrix / (2 * np.pi)


def _calculate_spec_matrix_helper(f, dim, var_coeffs, vma_coeffs, sigma):
    """
    Helper function to calculate spectral matrix for a single frequency.

    Args:
        f (float): Single frequency value.
        var_coeffs (np.ndarray): VAR coefficient array.
        vma_coeffs (np.ndarray): VMA coefficient array.
        sigma (np.ndarray): Covariance matrix or scalar variance.

    Returns:
        np.ndarray: Calculated spectral matrix for the given frequency.
    """
    if sigma.shape[0] == 1:
        cov_matrix = np.identity(dim) * sigma
    else:
        cov_matrix = sigma

    k_ar = np.arange(1, var_coeffs.shape[0] + 1)
    A_f_re_ar = np.sum(
        var_coeffs * np.cos(np.pi * 2 * k_ar * f)[:, np.newaxis, np.newaxis],
        axis=0,
    )
    A_f_im_ar = -np.sum(
        var_coeffs * np.sin(np.pi * 2 * k_ar * f)[:, np.newaxis, np.newaxis],
        axis=0,
    )
    A_f_ar = A_f_re_ar + 1j * A_f_im_ar
    A_bar_f_ar = np.identity(dim) - A_f_ar
    H_f_ar = np.linalg.inv(A_bar_f_ar)

    k_ma = np.arange(vma_coeffs.shape[0])
    A_f_re_ma = np.sum(
        vma_coeffs * np.cos(np.pi * 2 * k_ma * f)[:, np.newaxis, np.newaxis],
        axis=0,
    )
    A_f_im_ma = -np.sum(
        vma_coeffs * np.sin(np.pi * 2 * k_ma * f)[:, np.newaxis, np.newaxis],
        axis=0,
    )
    A_f_ma = A_f_re_ma + 1j * A_f_im_ma
    A_bar_f_ma = A_f_ma
    H_f_ma = A_bar_f_ma

    spec_mat = H_f_ar @ H_f_ma @ cov_matrix @ H_f_ma.conj().T @ H_f_ar.conj().T
    return spec_mat



import itertools
from typing import List, Optional

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import median_abs_deviation


def plot_triangle_psd_coherence(
    psd_q: np.ndarray,
    freqs: np.ndarray,
    true_psd: Optional[np.ndarray] = None,
    channel_labels: Optional[List[str]] = None,
    **kwargs,
) -> np.ndarray:
    """
    Create a triangle plot with PSDs on diagonal and coherence on off-diagonal.

    Parameters
    ----------
    psd_q : np.ndarray
        Quantiles of estimated PSDs with shape (n_quantiles, n_freqs, p, p)
    freqs : np.ndarray
        Frequency vector
    true_psd : np.ndarray, optional
        True PSD for comparison
    channel_labels : list, optional
        Labels for channels (e.g., ['H1', 'L1', 'V1'])

    Returns
    -------
    np.ndarray
        Array of axes objects
    """
    p = psd_q.shape[-1]

    # Create figure and axes
    fig, axs = plt.subplots(p, p, figsize=(p * 3, p * 2.5))
    if p == 1:
        axs = np.array([[axs]])

    # Plot diagonal elements (PSDs)
    for i in range(p):
        ax = axs[i, i]
        _plot_diagonal_psd(psd_q, freqs, i, ax, true_psd, **kwargs)

    # Plot off-diagonal elements (coherence)
    for i in range(p):
        for j in range(p):
            if i != j:
                ax = axs[i, j]
                _plot_off_diagonal_coherence(psd_q, freqs, i, j, ax, **kwargs)

    # Format the plot
    _format_triangle_plot(axs, channel_labels, **kwargs)

    return axs


def _plot_diagonal_psd(psd_q, freqs, idx, ax, true_psd=None, **kwargs):
    """Plot PSD on diagonal elements."""
    psd_ii = psd_q[..., idx, idx]

    # Default plotting parameters
    color = kwargs.get("color", "C0")

    # Plot quantiles if available
    if psd_q.shape[0] > 1:
        ax.fill_between(
            freqs, psd_ii[0], psd_ii[2], alpha=0.3, color=color, lw=0
        )
        ax.plot(freqs, psd_ii[1], color=color, lw=1.5)
    else:
        ax.plot(freqs, psd_ii.ravel(), color=color, lw=1.5)

    # Plot true PSD if provided
    if true_psd is not None:
        ax.plot(freqs, true_psd[..., idx, idx], "k--", alpha=0.7, lw=1)

    # Set log scale for PSD
    ax.set_yscale("log")
    ax.set_xscale("log")


def _plot_off_diagonal_coherence(psd_q, freqs, i, j, ax, **kwargs):
    """Plot coherence on off-diagonal elements."""
    # Compute coherence
    pxx = psd_q[..., i, i]
    pyy = psd_q[..., j, j]
    pxy = psd_q[..., i, j]

    coh_q, coh_uniform = compute_coherence(pxx, pyy, pxy)

    # Default color
    color = kwargs.get("coherence_color", "tab:blue")

    # Plot coherence
    if len(coh_q.shape) > 1:
        # Plot confidence bands
        ax.fill_between(
            freqs, coh_q[0], coh_q[2], alpha=0.3, color=color, lw=0
        )
        if coh_uniform is not None:
            ax.fill_between(
                freqs,
                coh_uniform[0],
                coh_uniform[2],
                alpha=0.5,
                color=color,
                lw=0,
            )
        # Plot median
        ax.plot(freqs, coh_q[1], color=color, lw=1.5)
    else:
        ax.plot(freqs, coh_q, color=color, lw=1.5)

    # Set coherence limits
    ax.set_ylim(0, 1)
    ax.set_xscale("log")


def _format_triangle_plot(axs, channel_labels=None, **kwargs):
    """Format the triangle plot."""
    p = axs.shape[0]

    # Default channel labels
    if channel_labels is None:
        channel_labels = [f"Ch{i + 1}" for i in range(p)]

    # Remove spines and ticks for upper triangle
    for i in range(p):
        for j in range(p):
            ax = axs[i, j]

            if i < j:  # Upper triangle - remove these plots
                ax.set_visible(False)
                continue

            # Format remaining plots
            if i == j:  # Diagonal
                # Add channel label
                label = f"${channel_labels[i]}$ PSD"
                ax.text(
                    0.05,
                    0.95,
                    label,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    fontweight="bold",
                )
                ax.set_ylabel("PSD [Hz$^{-1}$]")
            else:  # Lower triangle
                # Add coherence label
                label = f"$C_{{{channel_labels[i]}{channel_labels[j]}}}$"
                ax.text(
                    0.05,
                    0.95,
                    label,
                    transform=ax.transAxes,
                    verticalalignment="top",
                    fontweight="bold",
                )
                ax.set_ylabel("Coherence")

            # X-axis labels only on bottom row
            if i == p - 1:
                ax.set_xlabel("Frequency [Hz]")
            # else:
            # ax.set_xticklabels([])

    # Adjust layout
    plt.tight_layout()


def matrix_combinations(p):
    indices = range(p)
    combinations = itertools.combinations(indices, 2)
    result = [((i, j), (i, i), (j, j)) for i, j in combinations]
    return result


def uniformmax_help(sample):
    return np.abs(sample - np.median(sample)) / median_abs_deviation(sample)


def uniformmax_multi(coh_all):
    N_sample, N = coh_all.shape
    C_help = np.zeros((N_sample, N))
    for j in range(N):
        C_help[:, j] = uniformmax_help(coh_all[:, j])
    return np.max(C_help, axis=0)


def compute_coherence(pxx, pyy, pxy):
    coh_q = np.real(np.abs(pxy) ** 2) / np.real(pxx) / np.real(pyy)
    coh_uniform = None
    if len(coh_q.shape) > 1:
        coh_all = coh_q
        coh_q = np.quantile(coh_all, [0.05, 0.5, 0.95], axis=0)

        coh_median = coh_q[1]
        mad = median_abs_deviation(coh_all, axis=0, nan_policy="omit")
        mad[mad == 0] = 1e-10
        max_std_abs_dev = uniformmax_multi(coh_all)
        threshold = np.quantile(max_std_abs_dev, 0.9)
        coh_lower = coh_median - threshold * mad
        coh_upper = coh_median + threshold * mad
        coh_uniform = np.stack([coh_lower, coh_median, coh_upper], axis=0)

    return coh_q, coh_uniform



import h5py
import numpy as np

data = []
CHANNELS = ["X", "Y", "Z"]
for i, channel in enumerate(CHANNELS):
    with h5py.File(f"{channel}_ETnoise_GP.hdf5", "r") as f:
        data.append(f[f"E{i+1}:STRAIN"][:])
time = np.linspace(0, 2000, len(data[0]))
data = np.array(data)
data = data[:, : 2**19]
time = time[: 2**19]
with h5py.File("et_data.h5", "w") as f:
    # Create datasets for X, Y, Z
    f.create_dataset("X", data=data[0])
    f.create_dataset("Y", data=data[1])
    f.create_dataset("Z", data=data[2])
    f.create_dataset("time", data=time)



import warnings

import matplotlib.pyplot as plt
import numpy as np
from typing import List, Optional

warnings.filterwarnings(
    "ignore", message="Attempt to set non-positive ylim on a log-scaled axis"
)


def plot_psd(
        psdq: List[np.ndarray],
        true_psd: Optional[List[np.ndarray]] = None,
        pdgrm: Optional[List[np.ndarray]] = None,
        axes=None,
        **kwargs
) -> np.ndarray[plt.Axes]:
    """
    This is a utility function to plot the estimated PSDs.

    Parameters
    ----------
    psdq : tuple
        A tuple containing the estimated PSDs and the frequency vector.
    pdgrm : tuple, optional
        A tuple containing the periodogram and the frequency vector.
    true_psd : tuple, optional
        A tuple containing the true PSD and the frequency vector.
    **kwargs : dict
        Additional arguments to pass to the plotting functions.

    Returns
    -------
    np.ndarray[plt.Axes]
    """
    axes = plot_psdq(*psdq, **kwargs, axs=axes)
    if pdgrm:
        axes = plot_peridogram(*pdgrm, axs=axes, **kwargs)
    if true_psd:
        plot_single_psd(*true_psd, axes, **kwargs)
    format_axes(axes, **kwargs)
    return axes


def plot_psdq(psd_q, freqs, axs=None, **kwargs):
    """
    This is a utility class to plot the estimated multivariate PSDs.
    """
    nquantiles = psd_q.shape[0]
    nfreqs = psd_q.shape[1]
    p = psd_q.shape[-1]

    # setting some default values
    plt_kwargs = dict(
        color=kwargs.get("color", "C0"),
        zorder=kwargs.get("zorder", 10),
        ls=kwargs.get('ls', '-')
    )

    if nfreqs != len(freqs):
        raise ValueError(
            f"The len of frequencies {len(freqs)} does not match the len of PSD {nfreqs}"
        )

    if axs is None:
        # generate a square figure with pxp subplots
        fig, axs = _generate_fig(p)

    fill_kwargs = plt_kwargs.copy()
    fill_kwargs["alpha"] = 0.3
    fill_kwargs["lw"] = 0

    for row_i in range(p):
        for col_j in range(p):
            psd_ij = psd_q[..., row_i, col_j]
            psd_ij = _fmt_ij_elements(psd_ij, row_i, col_j)

            ax = axs[row_i, col_j]
            if nquantiles > 1:
                ax.fill_between(freqs, psd_ij[0], psd_ij[2], **fill_kwargs)
                ax.plot(freqs, psd_ij[1], **plt_kwargs)
            else:
                ax.plot(freqs.ravel(), psd_ij.ravel(), **plt_kwargs)

    return axs


def plot_psd_uniform(psd_uniform, freqs, axs=None, **kwargs):
    return plot_psdq(psd_uniform, freqs, axs, **kwargs)


def plot_single_psd(psd, freqs, axs=None, **kwargs):
    psd = np.array([psd])
    kwargs["color"] = kwargs.get("color", "k")
    kwargs["zorder"] = kwargs.get("zorder", -1)
    return plot_psdq(psd, freqs, axs, **kwargs)


def plot_peridogram(pdgrm, freq, axs=None, **kwargs):
    """
    This is a utility function to plot the periodogram of a time series.
    """
    p = pdgrm.shape[1]

    # setting some default values
    plt_kwargs = dict(
        color="lightgray",
        zorder=kwargs.get("zorder", -10),
        alpha=kwargs.get("alpha", 0.5),
    )

    if axs is None:
        fig, axs = _generate_fig(p)

    for row_i in range(p):
        for col_j in range(p):
            psd_ij = pdgrm[..., row_i, col_j]
            psd_ij = _fmt_ij_elements(psd_ij, row_i, col_j)

            ax = axs[row_i, col_j]
            ax.plot(freq, psd_ij, **plt_kwargs)

    if len(kwargs) > 0:
        format_axes(axs, **kwargs)

    return axs


def _fmt_ij_elements(psd, i, j):
    """
    This is a utility function to get the real and imaginary parts of the PSD.
    """
    if i > j:  # lower triangular
        return np.imag(psd)
    else:  # diag and upper triangular
        return np.real(psd)


def _generate_fig(p):
    """
    This is a utility function to generate a figure with pxp subplots.
    """
    fig, axs = plt.subplots(p, p, figsize=(p * 2.2, p * 2.2), sharex=True)
    return fig, axs


def format_axes(axes, **kwargs):
    """
    This is a utility function to format the figure.
    """
    _format_spines(axes, **kwargs)
    _format_text(axes, **kwargs)


def _format_spines(
        axes,
        tick_ln=5,
        diag_spline_thickness=2,
        xlims=None,
        diag_ylims=None,
        off_ylims=None,
        diag_log=True,
        off_symlog=True,
        sylmog_thresh=1e-49,
        **kwargs,
):
    p = axes.shape[0]
    plt.subplots_adjust(hspace=0.0, wspace=0.0)

    if diag_ylims is None:
        diag_ylims = (
            min([axes[i, i].get_ylim()[0] for i in range(p)]),
            max([axes[i, i].get_ylim()[1] for i in range(p)]),
        )

    if off_ylims is None:
        off_ylims = (
            min(
                [
                    axes[i, j].get_ylim()[0]
                    for i in range(p)
                    for j in range(p)
                    if i != j
                ]
            ),
            max(
                [
                    axes[i, j].get_ylim()[1]
                    for i in range(p)
                    for j in range(p)
                    if i != j
                ]
            ),
        )

    if xlims is None:
        xlims = axes[0, 0].get_xlim()

    # formatting for spline axes
    for i in range(p):
        for j in range(p):
            ax = axes[i, j]
            # turn off minor ticks
            ax.tick_params(
                "both",
                length=0,
                width=0,
                which="minor",
                bottom=False,
                top=False,
                left=False,
                right=False,
            )
            ax.tick_params(axis="y", direction="in", pad=-10)
            ax.yaxis.set_tick_params(which="both", labelleft=True, zorder=3)
            for label in ax.get_yticklabels():
                label.set_horizontalalignment("left")
            ax.set_xlim(xlims)

            if i == j:
                # increase ax-spine linewidth
                for spine in ax.spines.values():
                    spine.set_linewidth(diag_spline_thickness)
                    spine.set_zorder(10)
                ax.tick_params(
                    "both",
                    length=tick_ln,
                    width=diag_spline_thickness,
                    which="major",
                )
                if diag_log:
                    ax.set_yscale("log")
                ax.set_ylim(diag_ylims)
            else:
                ax.set_ylim(off_ylims)
                ax.patch.set_color("lightgray")  # or whatever color you like
                ax.patch.set_alpha(0.3)
                ax.tick_params("both", length=tick_ln, width=1, which="major")

                if off_symlog:
                    ax.set_yscale("symlog", linthresh=sylmog_thresh)


def _format_text(axes, channel_labels=None, **kwargs):
    p = axes.shape[0]
    if channel_labels is None:
        channel_labels = "".join([f"{i + 1}" for i in range(p)])
    assert len(channel_labels) == p

    add_chnl_lbls = kwargs.get("add_channel_labels", True)
    if add_chnl_lbls:
        # formatting text associated with the plot
        for i in range(p):
            for j in range(p):
                ax = axes[i, j]
                lbl = f"{channel_labels[i]}{channel_labels[j]}"
                lbl = "\mathbf{S}_{" + lbl + "}"

                if i < j:  # upper triangular
                    lbl = "$\Re(" + lbl + ")$"
                elif i > j:  # lower triangular
                    lbl = "$\Im(" + lbl + ")$"
                else:
                    lbl = "$" + lbl + "$"

                ax.text(
                    0.95,
                    0.95,
                    lbl,
                    transform=ax.transAxes,
                    horizontalalignment="right",
                    verticalalignment="top",
                    fontsize="small",
                )

    add_axes_labels = kwargs.get("add_axes_labels", True)
    if add_axes_labels:
        fig = axes[0, 0].get_figure()
        # yaxis label
        fig.text(0.075, 0.5, "PSD [1/Hz]", va="center", rotation="vertical")
        # xaxis label
        fig.text(0.5, 0.0325, "Frequency [Hz]", ha="center")



import matplotlib.pyplot as plt


def plot_losses(map_losses, kdl_losses, map_timing, kdl_timing):
    """
    Plot the posterior MAP values and KDL losses.

    Args:
        map_losses (list): List of posterior losses.
        kdl_losses (list): List of KDL losses.
    """
    fig, axes = plt.subplots(2,1, sharex=True)
    axes[0].plot(kdl_losses)
    axes[0].set_ylabel(r"$p(\theta | x)$")
    axes[1].plot(map_losses)
    axes[1].set_ylabel(r"$D_{KL}$")

    for ax in axes:
        ax.ticklabel_format(axis="y", style="sci", scilimits=(0, 0), )


    axes[1].set_xlabel("Iteration")
    axes[1].text(0.05, 0.8, f"Phase 2: {kdl_timing:.2f}s", transform=axes[1].transAxes, fontsize=10)
    axes[0].text(0.05, 0.8, f"Phase 1: {map_timing:.2f}s", transform=axes[0].transAxes, fontsize=10)
    plt.subplots_adjust(hspace=0)
    return axes




"""Module to compute the spectral density given the best surrogate posterior parameters and samples"""
import numpy as np
import tensorflow as tf
from typing import Tuple, List
from scipy.stats import median_abs_deviation

def compute_psd(
        Xmat_delta:tf.Tensor,
        Xmat_theta:tf.Tensor,
        p_dim:int,
        vi_samples: List[tf.Tensor],
        quantiles=[0.05, 0.5, 0.95],
        psd_scaling=1.0,
        fs=None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    This function is used to compute the spectral density given the best surrogate posterior parameters
    :param vi_samples: the surrogate posterior parameters

    Computes:
        1  psd_all: Nsamp instances of the spectral density [Nsamp, n-freq, dim, dim]
        2. pointwise_ci: the quantiles of the spectral density [n-quantiles, n-freq, dim, dim]
        3. uniform_ci: the uniform confidence interval of the spectral density [n-quantiles, n-freq, dim, dim]

    """
    delta2_all_s = tf.exp(
        tf.matmul(Xmat_delta, tf.transpose(vi_samples[0], [0, 2, 1]))
    )  # (500, #freq, p)

    theta_re_s = tf.matmul(
        Xmat_theta, tf.transpose(vi_samples[2], [0, 2, 1])
    )  # (500, #freq, p(p-1)/2)
    theta_im_s = tf.matmul(Xmat_theta, tf.transpose(vi_samples[4], [0, 2, 1]))

    theta_all_s = -(
        tf.complex(theta_re_s, theta_im_s)
    )  # (500, #freq, p(p-1)/2)
    theta_all_np = theta_all_s.numpy()

    D_all = tf.map_fn(
        lambda x: tf.linalg.diag(x), delta2_all_s
    ).numpy()  # (500, #freq, p, p)

    num_slices, num_freq, num_elements = theta_all_np.shape
    row_indices, col_indices = np.tril_indices(p_dim, k=-1)
    diag_matrix = np.eye(p_dim, dtype=np.complex64)
    T_all = np.tile(diag_matrix, (num_slices, num_freq, 1, 1))
    T_all[:, :, row_indices, col_indices] = theta_all_np.reshape(
        num_slices, num_freq, -1
    )

    T_all_conj_trans = np.conj(np.transpose(T_all, axes=(0, 1, 3, 2)))
    D_all_inv = np.linalg.inv(D_all)

    spectral_density_inverse_all = T_all_conj_trans @ D_all_inv @ T_all
    psd_all = np.linalg.inv(spectral_density_inverse_all)

    pointwise_ci = __get_pointwise_ci(psd_all, quantiles)
    uniform_ci = __get_uniform_ci(psd_all, pointwise_ci)

    # changing freq from [0, 1/2] to [0, samp_freq/2] (and applying scaling)
    if fs:
        original_fmax = 0.5
        true_fmax = fs / 2
        new_scale = true_fmax / original_fmax
        psd_all = psd_all / new_scale
        pointwise_ci = pointwise_ci / new_scale
        uniform_ci = uniform_ci / new_scale

    return (
        psd_all,
        pointwise_ci * psd_scaling ** 2,
        uniform_ci * psd_scaling ** 2,
    )


def __get_pointwise_ci(psd_all, quantiles):
    _, num_freq, p_dim, _ = psd_all.shape
    psd_q = np.zeros((3, num_freq, p_dim, p_dim), dtype=complex)

    diag_indices = np.diag_indices(p_dim)
    psd_q[:, :, diag_indices[0], diag_indices[1]] = np.quantile(
        np.real(psd_all[:, :, diag_indices[0], diag_indices[1]]),
        quantiles,
        axis=0,
    )

    # we dont do lower triangle because it is symmetric
    upper_triangle_idx = np.triu_indices(p_dim, k=1)
    real_part = np.real(psd_all[:, :, upper_triangle_idx[1], upper_triangle_idx[0]])
    imag_part = np.imag(psd_all[:, :, upper_triangle_idx[1], upper_triangle_idx[0]])

    for i, q in enumerate(quantiles):
        psd_q[i, :, upper_triangle_idx[1], upper_triangle_idx[0]] = (
                np.quantile(real_part, q, axis=0)
                + 1j * np.quantile(imag_part, q, axis=0)
        ).T

    psd_q[:, :, upper_triangle_idx[0], upper_triangle_idx[1]] = np.conj(
        psd_q[:, :, upper_triangle_idx[1], upper_triangle_idx[0]]
    )
    return psd_q


# Convert a complex matrix to a real matrix
def __complex_to_real(matrix):
    n = matrix.shape[0]
    real_matrix = np.zeros_like(matrix, dtype=float)
    real_matrix[np.triu_indices(n)] = np.real(matrix[np.triu_indices(n)])
    real_matrix[np.tril_indices(n, -1)] = np.imag(matrix[np.tril_indices(n, -1)])

    return real_matrix


# Find the normalized median absolute deviation for every element among all smaples
# For all samples of each frequency, each matrix, return their maximum normalized absolute deviation
def __uniformmax_multi(mSample):
    N_sample, N, d, _ = mSample.shape
    C_help = np.zeros((N_sample, N, d, d))

    for j in range(N):
        for r in range(d):
            for s in range(d):
                C_help[:, j, r, s] = __uniformmax_help(mSample[:, j, r, s])

    return np.max(C_help, axis=0)


def __uniformmax_help(sample):
    return np.abs(sample - np.median(sample)) / median_abs_deviation(sample)


# Convert a real matrix to a complex matrix
def __real_to_complex(matrix):
    n = matrix.shape[0]
    complex_matrix = np.zeros((n, n), dtype=complex)

    complex_matrix[np.diag_indices(n)] = matrix[np.diag_indices(n)]

    complex_matrix[np.triu_indices(n, 1)] = matrix[np.triu_indices(n, 1)] - 1j * matrix[np.tril_indices(n, -1)]
    complex_matrix[np.tril_indices(n, -1)] = matrix[np.triu_indices(n, 1)] + 1j * matrix[np.tril_indices(n, -1)]

    return complex_matrix


def __get_uniform_ci(psd_all, psd_q):
    psd_median = psd_q[1]
    n_samples, n_freq, p, _ = psd_all.shape

    # transform elements of psd_all and psd_median to the real numbers
    real_psd_all = np.zeros_like(psd_all, dtype=float)
    real_psd_median = np.zeros_like(psd_median, dtype=float)

    for i in range(n_samples):
        for j in range(n_freq):
            real_psd_all[i, j] = __complex_to_real(psd_all[i, j])

    for j in range(n_freq):
        real_psd_median[j] = __complex_to_real(psd_median[j])

    # find the maximum normalized absolute deviation for real_psd_all
    max_std_abs_dev = __uniformmax_multi(real_psd_all)
    # find the threshold of the 90% as a screening criterion
    threshold = np.quantile(max_std_abs_dev, 0.9)

    # find the uniform CI for real_psd_median
    mad = median_abs_deviation(real_psd_all, axis=0, nan_policy='omit')
    mad[mad == 0] = 1e-10
    lower_bound = real_psd_median - threshold * mad
    upper_bound = real_psd_median + threshold * mad

    # Converts lower_bound and upper_bound to the complex matrix
    psd_uni_lower = np.zeros_like(lower_bound, dtype=complex)
    psd_uni_upper = np.zeros_like(upper_bound, dtype=complex)

    for i in range(n_freq):
        psd_uni_lower[i] = __real_to_complex(lower_bound[i])
        psd_uni_upper[i] = __real_to_complex(upper_bound[i])

    psd_uniform = np.stack([psd_uni_lower, psd_median, psd_uni_upper], axis=0)

    return psd_uniform



import os

import h5py
import matplotlib.pyplot as plt
import numpy as np
import time

from sgvb_psd.psd_estimator import PSDEstimator

HERE = os.path.dirname(os.path.abspath(__file__))
DOCS_DIR = os.path.join(HERE, "..", "docs")
FNAME = os.path.join(DOCS_DIR, "examples", "et_data.h5")


def load_et_data(npts):
    with h5py.File(FNAME, "r") as f:
        channels = np.column_stack(
            [f["X"][:npts], f["Y"][:npts], f["Z"][:npts]]
        )
        return channels


def test_et(plot_dir):
    t0 = time.time()
    optim = PSDEstimator(
        x=load_et_data(2**14),
        N_theta=100,
        nchunks=8,
        ntrain_map=50,
        fs=2048,
        fmax_for_analysis=128,
        seed=0,
    )
    optim.run(lr=0.001)

    # run done, lets make some plots
    kwargs = dict(
        channel_labels="XYZ",
        sylmog_thresh=1e-49,
        xlims=[5, 128],
    )
    optim.plot(**kwargs, plot_periodogram=True)
    plt.savefig(f"{plot_dir}/ET_psd.png")
    optim.plot_coherence(labels="XYZ")
    plt.savefig(f"{plot_dir}/ET_coherence.png")

    end_time = time.time()
    estimation_time = end_time - t0
    assert estimation_time < 50, f"Estimation time {estimation_time} is too long"
    print(f"Test passed in {estimation_time} seconds")


from typing import Tuple

import numpy as np
import tensorflow as tf

from ..logging import logger


class AnalysisData:  # Parent used to create BayesianModel object
    def __init__(
        self,
        x: np.ndarray,
        nchunks: int = 128,
        fmax_for_analysis: float = 128,
        fs: float = 2048.0,
        N_theta: int = 15,
        N_delta: int = 15,
        fmin_for_analysis: float = None,
    ):
        # x:      N-by-p, multivariate timeseries with N samples and p dimensions
        # y_ft:   fourier transformed time series
        # freq:   frequencies w/ y_ft
        # p:  dimension of x
        # Xmat:   basis matrix
        # Zar:    arry of design matrix Z_k for every freq k
        self.x = x
        if x.shape[1] < 2:
            raise Exception("Time series should be at least 2 dimensional.")
        self.p = x.shape[1]
        self.nchunks = nchunks
        self.N_theta = N_theta
        self.N_delta = N_delta

        self.fs = fs
        self.fmax_for_analysis = fmax_for_analysis
        self.fmin_for_analysis = fmin_for_analysis

        # Compute the required datasets
        self.y_ft, self.freq = compute_chunked_fft(
            self.x,
            self.nchunks,
            self.fmax_for_analysis,
            self.fs,
            self.fmin_for_analysis,
        )
        self.Zar = _compute_chunked_Zmatrix(self.y_ft)
        Xmat_delta, Xmat_theta = _compute_Xmatrices(
            self.freq, N_delta, N_theta
        )

        # Setup tensors
        y_ft = tf.convert_to_tensor(self.y_ft, dtype=tf.complex64)
        self.y_re = tf.math.real(y_ft)
        self.y_im = tf.math.imag(y_ft)
        self.Xmat_delta = tf.convert_to_tensor(Xmat_delta, dtype=tf.float32)
        self.Xmat_theta = tf.convert_to_tensor(Xmat_theta, dtype=tf.float32)

        Zar = tf.convert_to_tensor(self.Zar, dtype=tf.complex64)
        self.Z_re = tf.math.real(Zar)
        self.Z_im = tf.math.imag(Zar)

        logger.info(f"Loaded {self}")

    def __repr__(self):
        x = self.x.shape
        y = self.y_ft.shape
        Xd = self.Xmat_delta.shape
        Xt = self.Xmat_theta.shape
        Z = self.Zar.shape
        return f"AnalysisData(x(t)={x}, y(f)={y}, Xmat_delta={Xd}, Xmat_theta={Xt}, Z={Z})"


def _compute_Xmatrices(
    freq, N_delta: int = 15, N_theta: int = 15
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Returns the X matrices for delta and theta based on the provided frequencies.

    Parameters:
    freq (np.ndarray): vector of frequencies
    N_delta (int): The number of basis functions to use for delta (default is 15).
    N_theta (int): The number of basis functions to use for theta (default is 15).

    Returns:
    Xd (np.ndarray): The design matrix for Demmler-Reinsch basis functions of delta,
                     the shape is (n, 2 + N_delta).
    Xt (np.ndarray): The design matrix for Demmler-Reinsch basis functions of theta,
                     the shape is (n, 2 + N_theta).

    """
    fstack = np.column_stack([np.repeat(1, freq.shape[0]), freq])
    Xd = np.concatenate([fstack, DR_basis(freq, N=N_delta)], axis=1)
    Xt = np.concatenate([fstack, DR_basis(freq, N=N_theta)], axis=1)
    return Xd, Xt


def _compute_Zmatrix(y_k: np.ndarray) -> np.ndarray:
    """
    Compute the design matrix Z_k for each frequency k.

    Parameters:
    y_k (np.ndarray): Fourier transformed time series data of shape (n, p).

    Returns:
    np.ndarray: Design matrix Z_k of shape (n, p, p*(p-1)/2).
    """
    n, p = y_k.shape
    Z_k = np.zeros((n, p, int(p * (p - 1) / 2)), dtype=np.complex64)

    for j in range(n):
        count = 0
        for i in range(1, p):
            Z_k[j, i, count : count + i] = y_k[j, :i]
            count += i

    return Z_k


def _compute_chunked_Zmatrix(y_ft: np.ndarray) -> np.ndarray:
    """
    Compute the design matrix Z, a 3D array (The design matrix Z_k for every frequency k).

    Parameters:
    y_ft (np.ndarray): Fourier transformed time series data of shape (chunks, n_per_chunk, p).

    Returns:
    np.ndarray: 3D array of design matrices Z_k for each frequency k.
    """
    chunks, n_per_chunk, p = y_ft.shape
    if p == 1:
        return np.zeros((chunks, n_per_chunk, 0), dtype=np.complex64)

    if chunks == 1:
        y_ls = np.squeeze(y_ft, axis=0)
        Z_ = _compute_Zmatrix(y_ls)
    else:
        y_ls = np.squeeze(np.split(y_ft, chunks))
        Z_ = np.array([_compute_Zmatrix(x) for x in y_ls])

    return Z_


def DR_basis(freq: np.ndarray, N=10):
    """
    Return the basis matrix for the Demmler-Reinsch basis
    for linear smoothing splines (Eubank,1999)

            # freq: vector of frequencies
    # N:  amount of basis used
    # return a len(freq)-by-N matrix
    """
    return np.array(
        [
            np.sqrt(2) * np.cos(x * np.pi * freq * 2)
            for x in np.arange(1, N + 1)
        ]
    ).T


def compute_chunked_fft(
    x: np.ndarray,
    nchunks: int,
    fmax_for_analysis: float,
    fs: float,
    fmin_for_analysis: float = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Scaled fft and get the elements of freq = 1:[Nquist] (or 1:[fmax_for_analysis] if specified)
    discarding the rest of freqs
    """

    if np.any(np.mean(x, axis=0) != 0) or np.any(np.std(x, axis=0) != 1):
        logger.warning("Input data not standardised!")

    orig_n, p = x.shape
    if orig_n < p:
        raise ValueError(
            f"Number of samples {orig_n} is less than number of dimensions {p}."
        )
    # split x into chunks
    n_per_chunk = x.shape[0] // nchunks
    chunked_x = np.array(np.split(x[0 : n_per_chunk * nchunks, :], nchunks))
    assert chunked_x.shape == (nchunks, n_per_chunk, p)

    chunked_x = chunked_x - np.mean(chunked_x, axis=1, keepdims=True)

    # compute fft for each chunk
    y_ft = np.apply_along_axis(np.fft.fft, 1, chunked_x)
    #
    # y = []
    # for i in range(nchunks):
    #     y_fft = np.apply_along_axis(np.fft.fft, 0, chunked_x[i])
    #     y.append(y_fft)
    # y = np.array(y)

    # scale it
    y_ft = y_ft / np.sqrt(n_per_chunk)
    Ts = 1  # for VB backend we use Duration of 1.0 (rescale later)
    fq_y = np.fft.fftfreq(np.size(chunked_x, axis=1), Ts)
    ftrue_y = np.fft.fftfreq(n_per_chunk, d=1 / fs)

    # Truncate the FFT'd data
    if np.mod(n_per_chunk, 2) == 0:  # n is even
        idx = int(n_per_chunk / 2)
    else:  # n is odd
        idx = int((n_per_chunk - 1) / 2)

    y_ft = y_ft[:, 1:idx, :]
    fq_y = fq_y[1:idx]
    ftrue_y = ftrue_y[1:idx]

    if fmax_for_analysis is None:
        fmax_for_analysis = ftrue_y[-1]
    fmax_idx = np.searchsorted(ftrue_y, fmax_for_analysis)
    fmin_idx = 0
    if fmin_for_analysis is not None:
        fmin_idx = np.searchsorted(ftrue_y, fmin_for_analysis)
    y_ft = y_ft[:, fmin_idx:fmax_idx, :]
    fq_y = fq_y[fmin_idx:fmax_idx]
    return y_ft, fq_y



import numpy as np
from sgvb_psd.utils.sim_varma import SimVARMA
from sgvb_psd.psd_estimator import PSDEstimator
import timeit
from typing import Tuple
from tqdm.auto import trange
from sgvb_psd.logging import logger
import matplotlib.pyplot as plt

logger.setLevel("ERROR")

np.random.seed(0)

DATA_KWGS = dict(
    var_coeffs=np.array(
        [[[0.5, 0.0], [0.0, -0.3]], [[0.0, 0.0], [0.0, -0.5]]]
    ),
    vma_coeffs=np.array([[[1.0, 0.0], [0.0, 1.0]]]),
    sigma=np.array([[1.0, 0.9], [0.9, 1.0]]),
)

PSD_KWGS = dict(
    N_theta=50,
    nchunks=1,
    ntrain_map=1000,
    max_hyperparm_eval=1,
    fs=2 * np.pi,
)
OPTIM_LR = 0.003


def run_simulation(log2n: int = 5, nrep:int = 5) -> Tuple[float, float]:
    sim_data = SimVARMA(
        n_samples=2 ** log2n,
        **DATA_KWGS,
    )
    func = lambda: PSDEstimator(
        x=sim_data.data,
        **PSD_KWGS
    ).run(lr=OPTIM_LR)
    times = np.array([timeit.repeat(func, repeat=nrep, number=1)])
    return np.median(times), np.std(times)


def run_simulations(minLog2n: int = 3, maxLog2n: int = 4, fname='timings.txt', nrep:int=5) -> None:
    log2ns = np.arange(minLog2n, maxLog2n + 1)
    times = np.zeros((len(log2ns), 2))
    # run in reverse order to start with the largest n
    for i in trange(len(log2ns), desc="Running simulations"):
        times[i, :] = run_simulation(log2ns[i], nrep=nrep)

    np.savetxt(
        fname,
        np.array([log2ns, times[:, 0], times[:, 1]]).T,
        header="log2n median_time std_time",
        fmt="%d %.4f %.4f",
    )


def plot_timings(fname: str = 'timings.txt') -> None:
    data = np.loadtxt(fname)
    log2n, median_time, std_time = data.T
    n = 2 ** log2n

    plt_fname = fname.replace('.txt', '.png')
    fig, ax = plt.subplots()
    ax.fill_between(
        n,
        median_time - std_time,
        median_time + std_time,
        alpha=0.2,
        color="tab:blue",
    )
    ax.loglog(n, median_time, color="tab:blue")
    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.set_xlabel("n")
    ax.set_ylabel("Time (s)")
    plt.savefig(plt_fname, bbox_inches="tight")


def benchmark(minLog2n: int = 3, maxLog2n: int = 4, fname='timings.txt', nrep:int=5) -> None:
    run_simulations(minLog2n, maxLog2n, fname, nrep=nrep)
    plot_timings(fname)
